import os
import json

import numpy as np
import scipy.signal as signal
import pandas as pd

from .focal_mechanism import check_convert_fm
from .geo import rotate_rtz_to_enz
from .pytaup import read_tpts_table
from .utils import shift_green2real_tpts
from .signal_process import resample


def read_time_series_qseis06_bin(path_greenfunc, start_count):
    time_series_list = []
    for com in ["tr", "tz"]:
        time_series_com = np.load(os.path.join(path_greenfunc, "grn_%s.npy" % com))
        time_series_list.append(time_series_com[:, start_count].reshape(4, -1))
    for com in ["tt"]:
        time_series_com = np.load(os.path.join(path_greenfunc, "grn_%s.npy" % com))
        time_series_list.append(time_series_com[:, start_count].reshape(2, -1))
    return time_series_list


def read_time_series_qseis06_ascii(path_greenfunc, start_count):
    time_seris_list = []
    start_count = start_count + 1
    for com in ["tr", "tz"]:
        ex_com = pd.read_csv(
            os.path.join(path_greenfunc, "ex.%s" % com), sep="\\s+"
        ).to_numpy()  # type:ignore
        ss_com = pd.read_csv(
            os.path.join(path_greenfunc, "ss.%s" % com), sep="\\s+"
        ).to_numpy()  # type:ignore
        ds_com = pd.read_csv(
            os.path.join(path_greenfunc, "ds.%s" % com), sep="\\s+"
        ).to_numpy()  # type:ignore
        cl_com = pd.read_csv(
            os.path.join(path_greenfunc, "cl.%s" % com), sep="\\s+"
        ).to_numpy()  # type:ignore
        time_series_com = np.concatenate(
            [
                ex_com[:, start_count],
                ss_com[:, start_count],
                ds_com[:, start_count],
                cl_com[:, start_count],
            ]
        ).reshape(4, -1)
        time_series_com = np.array(time_series_com, dtype=np.float32)
        time_seris_list.append(time_series_com)

    for com in ["tt"]:
        ss_r = pd.read_csv(
            os.path.join(path_greenfunc, "ss.%s" % com), sep="\\s+"
        ).to_numpy()  # type:ignore
        ds_r = pd.read_csv(
            os.path.join(path_greenfunc, "ds.%s" % com), sep="\\s+"
        ).to_numpy()  # type:ignore
        time_series_com = np.concatenate(
            [
                ss_r[:, start_count],
                ds_r[:, start_count],
            ]
        ).reshape(2, -1)
        time_series_com = np.array(time_series_com, dtype=np.float32)
        time_seris_list.append(time_series_com)
    return time_seris_list


def synthesize_qseis06(time_series_list, m1, m2):
    r = (
        time_series_list[0][0] * m1[0]
        + time_series_list[0][1] * m1[1]
        + time_series_list[0][2] * m1[2]
        + time_series_list[0][3] * m1[3]
    )
    z = (
        time_series_list[1][0] * m1[0]
        + time_series_list[1][1] * m1[1]
        + time_series_list[1][2] * m1[2]
        + time_series_list[1][3] * m1[3]
    )
    t = time_series_list[2][0] * m2[0] + time_series_list[2][1] * m2[1]
    return r, t, z


def seek_qseis06(
    path_green,
    event_depth_km,
    receiver_depth_km,
    az_deg,
    dist_km,
    focal_mechanism,
    srate,
    output_type="disp",
    rotate=True,
    before_p=None,
    pad_zeros=False,
    shift=False,
    only_seismograms=True,
    model_name="ak135fc",
    green_info=None,
):
    if green_info is None:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            green_info = json.load(fr)
    srate_grn = 1 / green_info["sampling_interval"]
    sampling_num = green_info["sampling_num"]
    time_reduction_velo = green_info["time_reduction_velo"]
    dist_range = green_info["dist_range"]
    delta_dist = green_info["delta_dist"]
    num_each_group = green_info["N_each_group"]
    wavelet_type = green_info["wavelet_type"]
    grn_dep_list = green_info["event_depth_list"]
    grn_receiver_list = green_info["receiver_depth_list"]
    if not isinstance(grn_dep_list, list):
        grn_dep_source = grn_dep_list
    else:
        grn_dep_source = grn_dep_list[
            np.argmin(np.abs(event_depth_km - np.array(grn_dep_list)))
        ]
    if not isinstance(grn_receiver_list, list):
        grn_dep_receiver = grn_receiver_list
    else:
        grn_dep_receiver = grn_receiver_list[
            np.argmin(np.abs(receiver_depth_km - np.array(grn_receiver_list)))
        ]

    path_greenfunc = str(
        os.path.join(path_green, "%.2f" % grn_dep_source, "%.2f" % grn_dep_receiver)
    )
    ind = round((dist_km - dist_range[0]) / delta_dist)
    ind_group = ind // num_each_group
    grn_dist = dist_range[0] + ind * delta_dist
    start_count = ind - ind_group * num_each_group

    path_greenfunc_sub = os.path.join(path_greenfunc, "%d_0" % ind_group)
    if os.path.exists(os.path.join(path_greenfunc_sub, "grn_tz.npy")):
        time_series_list = read_time_series_qseis06_bin(
            path_greenfunc=path_greenfunc_sub, start_count=start_count
        )
    else:
        time_series_list = read_time_series_qseis06_ascii(
            path_greenfunc=path_greenfunc_sub, start_count=start_count
        )

    # r,t,z
    [M11, M12, M13, M22, M23, M33] = check_convert_fm(focal_mechanism=focal_mechanism)

    exp = (M11 + M22 + M33) / 3
    ss1 = M12
    ss2 = (M11 - M22) / 2
    ds1 = M13
    ds2 = M23
    clvd = (-M11 / 2 - M22 / 2 + M33) / (3 / 2)

    az_rad = np.deg2rad(az_deg)
    sin_az, cos_az = np.sin(az_rad), np.cos(az_rad)
    sin_2az, cos_2az = np.sin(2 * az_rad), np.cos(2 * az_rad)

    m1 = [exp, ss1 * sin_2az + ss2 * cos_2az, ds1 * cos_az + ds2 * sin_az, clvd]
    m2 = [ss1 * cos_2az - ss2 * sin_2az, ds1 * sin_az - ds2 * cos_az]

    r, t, z = synthesize_qseis06(time_series_list=time_series_list, m1=m1, m2=m2)

    if rotate:
        seismograms = rotate_rtz_to_enz(az_in_deg=az_deg, r=r, t=-t, z=-z)
    else:
        seismograms = np.array([r, -t, -z])

    tpts_table = None
    if (before_p is not None) or shift or pad_zeros:
        first_p_grn, first_s_grn = read_tpts_table(
            path_green=path_green,
            event_depth_km=grn_dep_source,
            receiver_depth_km=grn_dep_receiver,
            ind=ind,
        )
        tpts_table = {"p_onset": first_p_grn, "s_onset": first_s_grn}

    if time_reduction_velo != 0:
        time_reduction = grn_dist / time_reduction_velo
    else:
        time_reduction = 0

    ts_count = 0
    if before_p is not None:
        ts_count = round(
            (tpts_table["p_onset"] - time_reduction - before_p) * srate_grn
        )
    if pad_zeros:
        if before_p is not None:
            raise ValueError("can not set before_p and pad_zeros together")
        ts_count = round(-time_reduction * srate_grn)
    seismograms = np.roll(seismograms, -ts_count)
    if ts_count > 0:
        seismograms[:, -ts_count:] = 0
    elif ts_count < 0:
        seismograms[:, :-ts_count] = 0

    first_p = None
    first_s = None
    if shift:
        seismograms, first_p, first_s = shift_green2real_tpts(
            seismograms=seismograms,
            tpts_table=tpts_table,
            srate=srate_grn,
            green_before_p=tpts_table["p_onset"] - time_reduction,
            event_depth_km=event_depth_km,
            dist_in_km=dist_km,
            receiver_depth_km=receiver_depth_km,
            model_name=model_name,
        )

    seismograms_resample = np.zeros((3, round(sampling_num * srate / srate_grn)))
    for i in range(3):
        seismograms_resample[i] = resample(
            seismograms[i], srate_old=srate_grn, srate_new=srate, zero_phase=True
        )

    if wavelet_type == 1 and output_type == "disp":
        seismograms_resample = np.cumsum(seismograms_resample, axis=1) / srate
    elif wavelet_type == 2 and output_type == "velo":
        seismograms_resample = (
            signal.convolve(
                seismograms_resample.T,
                np.array([1, -1])[:, None],
                mode="same",
                method="auto",
            ).T
            / srate
        )
    if only_seismograms:
        return seismograms_resample
    else:
        return (
            seismograms_resample,
            tpts_table,
            first_p,
            first_s,
            grn_dep_source,
            grn_dep_receiver,
            grn_dist,
        )


if __name__ == "__main__":
    pass



import matplotlib

matplotlib.use('tkagg')
import matplotlib.pyplot as plt
from scipy.signal import firwin, freqz, lfilter
import pandas as pd
import numpy as np

from pygrnwang.read_qseis06 import *
from pygrnwang.read_qssp2020 import *
from pygrnwang.read_spgrn import *
from pygrnwang.signal_process import resample, filter_butter

if __name__ == '__main__':
    event_depth_km = 10
    receiver_depth_km = 1
    az_deg = 20
    dist_km = 300
    srate = 0.25
    fm = [30, 40, 50]
    tp,ts=cal_first_p_s(
        event_depth_km,dist_km,receiver_depth_km)
    print(tp, ts)

    ind = 2
    cutoff = 0.05
    cut_length = 64

    numtaps = 41
    taps = firwin(numtaps, cutoff, fs=srate, window='hamming')

    u_qseis = seek_qseis06(
        path_green="/e/grns_test/test_qseis",
        event_depth_km=event_depth_km,
        receiver_depth_km=receiver_depth_km,
        az_deg=az_deg,
        dist_km=dist_km,
        focal_mechanism=fm,
        srate=srate,
        rotate=True,
        before_p=20,
        pad_zeros=False,
        shift=False,
        only_seismograms=True,
        model_name="/e/grns_test/test_qseis/noq.nd",
        green_info=None,
    )
    u_qseis = u_qseis[ind]
    #u_qseis = u_qseis[2:]
    u_qseis_conv = np.convolve(u_qseis, taps, mode='full')

    u_qssp = seek_qssp2020(
        path_green="/e/grns_test/test_qssp",
        event_depth_km=event_depth_km,
        receiver_depth_km=receiver_depth_km,
        az_deg=az_deg,
        dist_km=dist_km,
        focal_mechanism=fm,
        srate=srate,
        before_p=20,
        pad_zeros=False,
        shift=False,
        rotate=True,
        only_seismograms=True,
        output_type="velo",
        model_name="/e/grns_test/test_qssp/noq.nd",
        green_info=None,
    )
    u_qssp = u_qssp[ind]
    u_qssp_conv = np.convolve(u_qssp, taps, mode='full')

    u_spgrn = seek_spgrn2020(
        path_green="/e/grns_test/test_spgrn",
        event_depth_km=event_depth_km,
        receiver_depth_km=receiver_depth_km,
        az_deg=az_deg,
        dist_km=dist_km,
        focal_mechanism=fm,
        srate=srate,
        before_p=20,
        pad_zeros=False,
        shift=False,
        rotate=True,
        only_seismograms=True,
        model_name="/e/grns_test/test_spgrn/noq.nd",
        green_info=None,
    )
    u_spgrn = u_spgrn[ind]
    u_spgrn_conv = np.convolve(u_spgrn, taps, mode='full')

    # plt.figure()
    # plt.plot(taps)
    # plt.show()

    plt.figure()
    plt.plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
             u_qseis[:cut_length], label='qseis')
    plt.plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
             u_qssp[:cut_length], label='qssp')
    plt.plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
             u_spgrn[:cut_length], label='spgrn')
    plt.legend()
    # plt.vlines(x=tp, ymin=np.min(u_qseis[:cut_length]),
    #            ymax=np.max(u_qseis[:cut_length]),color='red')
    plt.xlabel('Time (s)')
    plt.show()

    plt.figure()
    plt.plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
             u_qseis_conv[:cut_length], label='qseis')
    plt.plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
             u_qssp_conv[:cut_length], label='qssp')
    plt.plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
             u_spgrn_conv[:cut_length], label='spgrn')
    plt.legend()
    plt.vlines(x=tp, ymin=np.min(u_qseis_conv[:cut_length]),
               ymax=np.max(u_qseis_conv[:cut_length]),color='red')
    plt.xlabel('Time (s)')
    plt.show()





import numpy as np

from pygrnwang.create_edgrn_bulk import (
    pre_process_edgrn2,
    create_grnlib_edgrn2_parallel,
)
from pygrnwang.create_edcmp_bulk import (
    pre_process_edcmp2,
    compute_static_stress_edcmp2_parallel,
)

if __name__ == "__main__":
    path_green = "/e/grns_test/test_edcmp"
    path_nd = '/home/zjc/python_works/pygrnwang/test/ak135fc_layer_660.nd'
    path_bin_edgrn = "/home/zjc/python_works/pygrnwang/pygrnwang/exec/edgrn2.bin"
    path_bin_edcmp = "/home/zjc/python_works/pygrnwang/pygrnwang/exec/edcmp2.bin"

    pre_process_edgrn2(
        processes_num=8,
        path_green=path_green,
        path_bin=path_bin_edgrn,
        grn_source_depth_range=[0, 30],
        grn_source_delta_depth=0.5,
        grn_dist_range=[0, 200],
        grn_delta_dist=0.1,
        obs_depth_list=list(np.linspace(0, 30, 61)),
        wavenumber_sampling_rate=256,
        path_nd=path_nd,
        earth_model_layer_num=30,
    )
    create_grnlib_edgrn2_parallel(
        path_green=path_green, check_finished=False
    )

    # source_array = np.array([[0, 0, 10, 0.001, 0.001, 1, np.nan, 0, 90, 0]])
    # source_array_new = np.zeros((len(source_array), 9))
    # source_array_new[:, 0] = source_array[:, 8]
    # source_array_new[:, 1:9] = source_array[:, 0:8]
    # pre_process_edcmp2(
    #     processes_num=8,
    #     path_green=path_green,
    #     path_bin=path_bin_edcmp,
    #     obs_depth_list=[0, 10, 20],
    #     obs_x_range=[-3, 3],
    #     obs_y_range=[-3, 3],
    #     obs_delta_x=0.01,
    #     obs_delta_y=0.01,
    #     source_array=source_array_new,
    #     source_ref=[0, 0],
    #     obs_ref=[0, 0],
    #     layered=True,
    # )
    # compute_static_stress_edcmp2_parallel_single_node(
    #     path_green=path_green, check_finished=False
    # )



"""
pygrnwang
==========
This Python package serves as the frontend for calculating and building a Green's function library
for synthetic seismograms. The backend consists of Wang Rongjiang's program for calculating synthetic
seismograms in layered Earth models, including EDGRN/EDCMP QSEIS, SPGRN, and QSSP (Wang, 1999;
Wang et al., 2017).

"""

__version__ = "1.0.0"



import os

import numpy as np
import pandas as pd

from .geo import d2m
from .utils import convert_earth_model_nd2inp, call_exe

mt_com_list = ["mrr", "mtt", "mpp", "mrt", "mrp", "mtp"]
output_type_list = [
    "disp",
    "velo",
    "acce",
    "rota",
    "rota_rate",
    "stress",
    "stress_rate",
    "strain",
    "strain_rate",
    "gravitation",
    "gravimeter",
]


def create_dir_qssp2020(event_depth, receiver_depth, path_green):
    path_func = os.path.join(
        path_green, "GreenFunc", "%.2f" % event_depth, "%.2f" % receiver_depth, ""
    )
    os.makedirs(path_func, exist_ok=True)
    for mt_com in mt_com_list:
        os.makedirs(os.path.join(str(path_func), mt_com), exist_ok=True)

    path_spec = os.path.join(
        path_green, "GreenSpec", "%.2f" % event_depth, "%.2f" % receiver_depth, ""
    )
    os.makedirs(path_spec, exist_ok=True)
    return path_func, path_spec


def create_points(dist_range, delta_dist):
    """
    :param dist_range: km
    :param delta_dist: km
    """
    rmax = dist_range[1]
    rmin = dist_range[0]
    points_lat = np.linspace(rmin, rmax, round(np.ceil((rmax - rmin) / delta_dist)) + 1)
    points_lat = points_lat * 1e3 / d2m
    points = np.concatenate([points_lat, np.zeros_like(points_lat)])
    points = points.reshape((2, -1)).T
    return points


def create_locs(points, time_reduction):
    lines_locs = []
    for i in range(len(points)):
        loc = "%f %f 'Loc%d' %f\n" % (points[i, 0], points[i, 1], i, time_reduction)
        lines_locs.append(loc)
    return lines_locs


def create_inp_qssp2020(
    mt_com,
    path_green,
    event_depth,
    receiver_depth,
    spec_time_window,
    sampling_interval,
    max_frequency,
    max_slowness,
    anti_alias,
    turning_point_filter,
    turning_point_d1,
    turning_point_d2,
    free_surface_filter,
    gravity_fc,
    gravity_harmonic,
    cal_sph,
    cal_tor,
    min_harmonic,
    max_harmonic,
    source_radius,
    cal_gf,
    source_duration,
    output_observables: list,
    time_window,
    time_reduction,
    dist_range,
    delta_dist,
    path_nd=None,
    earth_model_layer_num=None,
    physical_dispersion=0,
):
    path_func = str(
        os.path.join(
            path_green,
            "GreenFunc",
            "%.2f" % event_depth,
            "%.2f" % receiver_depth,
            mt_com,
            "",
        )
    )
    path_spec = str(
        os.path.join(
            path_green,
            "GreenSpec",
            "%.2f" % event_depth,
            "%.2f" % receiver_depth,
            "",
        )
    )
    path_inp = os.path.join(path_green, "qssp2020.inp")
    if os.path.exists(path_inp):
        with open(path_inp, "r") as fr:
            lines = fr.readlines()
    else:
        from .qssp2020inp import s

        lines = s.split("\n")
        lines = [line + "\n" for line in lines]
    last_line = [lines[-1]]
    lines_earth_head = lines[141:155]
    lines_earth = lines[155:-1]
    lines = lines[:141]  # cutoff locs

    lines[24] = "%.2f\n" % receiver_depth

    lines[52] = "%f  %f\n" % (spec_time_window, sampling_interval)
    lines[53] = "%f\n" % max_frequency
    lines[54] = "%f\n" % max_slowness
    lines[55] = "%f\n" % anti_alias
    lines[56] = "%d %f %f\n" % (
        turning_point_filter,
        turning_point_d1,
        turning_point_d2,
    )
    lines[57] = "6371.0 %d\n" % free_surface_filter

    lines[65] = "%f %d\n" % (gravity_fc, gravity_harmonic)

    lines[75] = "%d %d %d %d\n" % (cal_sph, cal_tor, min_harmonic, max_harmonic)

    lines[86] = "1 %f '%s'\n" % (source_radius, path_spec)
    lines[87] = "%.2f 'Green_%.2fkm' %d\n" % (event_depth, event_depth, cal_gf)

    lines[111] = "1 1\n"
    mt = [0 for _ in range(6)]
    if mt_com in mt_com_list:
        ind_mt = mt_com_list.index(mt_com)
        mt[ind_mt] = 1
    lines[112] = (
        "1.0 "
        + " ".join("%f" % mt[_] for _ in range(6))
        + " 0.0 0.0 %.2f 0.0 %f\n" % (event_depth, source_duration)
    )
    lines[135] = " ".join(["%d" % output_observables[_] for _ in range(11)]) + "\n"
    lines[136] = "'%s'\n" % path_func
    lines[137] = "%f\n" % time_window
    lines[138] = "0 0 0\n"
    lines[139] = "0 %f\n" % max_slowness

    points = create_points(dist_range=dist_range, delta_dist=delta_dist)
    n_locs = len(points)
    lines[140] = "%s\n" % n_locs
    lines_locs = create_locs(points=points, time_reduction=time_reduction)
    lines = lines + lines_locs

    if path_nd is not None:
        lines_earth = convert_earth_model_nd2inp(
            path_nd=path_nd, path_output="earth_model.dat"
        )
    if earth_model_layer_num is None:
        earth_model_layer_num = len(lines_earth)
    lines_earth = lines_earth[:earth_model_layer_num]
    lines_earth_head[7] = "%d  %d\n" % (earth_model_layer_num, physical_dispersion)
    if mt_com == "spec":
        path_inp = os.path.join(
            path_spec,
            "spec.inp",
        )
    elif mt_com in mt_com_list:
        ind_mt = mt_com_list.index(mt_com)
        path_inp = os.path.join(
            path_func,
            "%s.inp" % mt_com_list[ind_mt],
        )
    else:
        return ValueError("mt_com wrong!!!")
    with open(path_inp, "w") as fw:
        fw.writelines(lines + lines_earth_head + lines_earth + last_line)
    return path_inp


def call_qssp2020(
    event_depth, receiver_depth, mt_com, path_green, check_finished=False
):
    os.chdir(path_green)
    if mt_com == "spec":
        path_inp = str(
            os.path.join(
                path_green,
                "GreenSpec",
                "%.2f" % event_depth,
                "%.2f" % receiver_depth,
                "spec.inp",
            )
        ).replace("'", "")
    elif mt_com in mt_com_list:
        path_inp = str(
            os.path.join(
                path_green,
                "GreenFunc",
                "%.2f" % event_depth,
                "%.2f" % receiver_depth,
                mt_com,
                "%s.inp" % mt_com,
            )
        ).replace("'", "")
    else:
        return ValueError("mt_com wrong!!!")

    sub_dir = os.path.dirname(path_inp)
    path_finished = os.path.join(sub_dir, ".finished")
    if (
        check_finished
        and os.path.exists(path_finished)
        and len(os.listdir(sub_dir)) > 2
    ):
        with open(path_finished, "r") as fr:
            output = fr.readlines()
        return output

    output = call_exe(
        path_green=path_green,
        path_inp=path_inp,
        path_finished=path_finished,
        name="qssp2020",
    )
    return output


def convert_pd2bin_qssp2020(path_green, event_depth, receiver_depth, output_type_ind):
    if output_type_list[output_type_ind] == "gravimeter":
        enz_list = [""]
    elif output_type_list[output_type_ind] in [
        "disp",
        "velo",
        "acce",
        "rota",
        "rota_rate",
        "gravitation",
    ]:
        enz_list = ["_e", "_n", "_z"]
    elif output_type_list[output_type_ind] in [
        "stress",
        "stress_rate",
        "strain",
        "strain_rate",
    ]:
        enz_list = ["_ee", "_en", "_ez", "_nn", "_nz", "_zz"]
    else:
        raise ValueError("output_type wrong")
    for k in range(6):
        for l in range(len(enz_list)):
            path_dat = str(
                os.path.join(
                    path_green,
                    "GreenFunc",
                    "%.2f" % event_depth,
                    "%.2f" % receiver_depth,
                    mt_com_list[k],
                    "_%s%s.dat" % (output_type_list[output_type_ind], enz_list[l]),
                )
            )
            dat = pd.read_csv(path_dat, sep="\\s+").to_numpy()
            dat = np.array(dat, dtype=np.float32)[:, 1:].T
            dat.tofile(os.path.join(path_dat[:-4], "%s.bin" % path_dat[:-4]))


if __name__ == "__main__":
    pass



import os
import json

import numpy as np
import pandas as pd

from .create_qssp2020 import mt_com_list
from .utils import (
    shift_green2real_tpts,
    create_rotate_z_mat,
    rotate_symmetric_tensor_series,
)
from .focal_mechanism import convert_mt_axis, mt2full_mt_matrix, check_convert_fm
from .geo import rotate_rtz_to_enz
from .signal_process import resample

# Define output type lists
one_com_list = ["gravimeter"]
three_com_list = ["disp", "velo", "acce", "rota", "rota_rate", "gravitation"]
six_com_list = ["strain", "strain_rate", "stress", "stress_rate"]


def read_time_series_qssp2020(path_bin, ind, sampling_num):
    fr = open(path_bin, "rb")
    time_series = np.fromfile(
        file=fr, dtype=np.float32, count=sampling_num, offset=ind * sampling_num * 4
    )
    fr.close()
    return time_series


def read_tpts_table_qssp2020(path_green, event_depth_km, receiver_depth_km, ind):
    fr_tp = open(
        os.path.join(
            path_green,
            "GreenFunc",
            "%.2f" % event_depth_km,
            "%.2f" % receiver_depth_km,
            "tp_table.bin",
        ),
        "rb",
    )
    tp = np.fromfile(file=fr_tp, dtype=np.float32, count=1, offset=ind * 4)[0]
    fr_tp.close()

    fr_ts = open(
        os.path.join(
            path_green,
            "GreenFunc",
            "%.2f" % event_depth_km,
            "%.2f" % receiver_depth_km,
            "ts_table.bin",
        ),
        "rb",
    )
    ts = np.fromfile(file=fr_ts, dtype=np.float32, count=1, offset=ind * 4)[0]
    fr_ts.close()
    return float(tp), float(ts)


def seek_raw_qssp2020(
    path_green,
    event_depth,
    receiver_depth,
    dist,
    output_type="disp",
    green_info=None,
):
    """
    read raw data at stations located on the straight line from the origin to the north
    moment tensor in rtp axis
    data in enz axis
    :param path_green:
    :param event_depth: km
    :param receiver_depth: km
    :param dist: km
    :param output_type: str, 'output_type must in  disp | velo | acce | strain | strain_rate | '
            'stress | stress_rate | rotation | rotation_rate | gravitation | gravimeter'
    :param green_info:
    """
    if green_info is None:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            green_info = json.load(fr)
    N_T = round(green_info["time_window"] / green_info["sampling_interval"]) + 1
    if output_type == "gravimeter":
        enz_list = [""]
        data_all = np.zeros((N_T, 6))
    elif output_type in ["disp", "velo", "acce", "rota", "rota_rate", "gravitation"]:
        enz_list = ["_e", "_n", "_z"]
        data_all = np.zeros((N_T, 18))
    elif output_type in ["stress", "stress_rate", "strain", "strain_rate"]:
        enz_list = ["_ee", "_en", "_ez", "_nn", "_nz", "_zz"]
        data_all = np.zeros((N_T, 36))
    else:
        raise ValueError(
            "output_type must in  disp | velo | acce | strain | strain_rate | "
            "stress | stress_rate | rota | rota_rate | gravitation | gravimeter"
        )

    for i_rtp in range(6):
        path_func = str(
            os.path.join(
                path_green,
                "GreenFunc",
                "%.2f" % event_depth,
                "%.2f" % receiver_depth,
                mt_com_list[i_rtp],
                "",
            )
        )
        for i_enz in range(len(enz_list)):
            dist_range = green_info["dist_range"]
            delta_dist = green_info["delta_dist"]
            ind = round((dist - dist_range[0]) / delta_dist)
            path_bin = os.path.join(
                path_func, "_%s%s.bin" % (output_type, enz_list[i_enz])
            )
            # path_bin = ''
            if os.path.exists(path_bin):
                data_enz = read_time_series_qssp2020(
                    path_bin=path_bin, ind=ind, sampling_num=N_T
                )
                data_all[:, i_enz + len(enz_list) * i_rtp] = data_enz
            else:
                data_enz = pd.read_csv(
                    os.path.join(
                        path_func, "_%s%s.dat" % (output_type, enz_list[i_enz])
                    ),
                    sep="\\s+",
                ).to_numpy()[:, 1:]
                data_all[:, i_enz + len(enz_list) * i_rtp] = data_enz[:, ind]
    return data_all


def seek_qssp2020(
    path_green,
    event_depth_km,
    receiver_depth_km,
    az_deg,
    dist_km,
    focal_mechanism,
    srate,
    before_p=None,
    pad_zeros=False,
    shift=False,
    rotate=True,
    only_seismograms=True,
    output_type="disp",
    model_name="ak135",
    green_info=None,
):
    """
    Read seismic data with either one, three, or six components based on output_type.

    Depending on output_type, the function reads:
      - one component (e.g., gravimeter),
      - three components (e.g., displacement, velocity, acceleration, rotation, rotation rate, gravitation),
      - or six components (e.g., strain, strain rate, stress, stress rate).

    Parameters:
        path_green (str): Root directory of the data.
        event_depth_km (float): Event depth in km.
        receiver_depth_km (float): Receiver depth in km.
        az_deg (float): Azimuth in degrees.
        dist_km (float): Epicentral distance in km.
        focal_mechanism (ndarray/list):
                strike, dip, rake(deg), length=3 or
                [M11, M12, M13, M22, M23, M33], in NED coordinates, length=6.
        srate (float): Sampling rate in Hz.
        before_p (float, optional): Time before the P-wave arrival in seconds.
        pad_zeros (bool, optional): Whether to pad with zeros.
        shift (bool, optional): Whether to shift the seismograms.
        rotate (bool, optional): Whether to perform rotation from rtz2ned.
        only_seismograms (bool, optional): If True, only return seismograms.
        output_type (str, optional): Output type string. For example, 'disp','velo','acce',
                                     'rota','rota_rate','gravimeter','gravitation',
                                     'strain','strain_rate','stress','stress_rate'.
        model_name (str, optional): Model name string. If not provided, a default is set based on output_type:
                                    - For one or three component types: "ak135"
                                    - For six component types: "ak135" (or adjust as needed)
        green_info (dict, optional): Information of Green's function libarary
    Returns:
        If only_seismograms is True:
            seismograms (ndarray): For one-component: (1 x N_T),
                                   for three-component: (3 x N_T),
                                   for six-component: (6 x N_T).
        Otherwise, returns a tuple:
            (seismograms, tpts_table, first_p, first_s, grn_dist)
            where tpts_table is a dict with keys "p_onset" and "s_onset".
    """
    if green_info is None:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            green_info = json.load(fr)
    srate_grn = 1 / green_info["sampling_interval"]
    sampling_num = (
        round(green_info["time_window"] / green_info["sampling_interval"]) + 1
    )
    dist_range = green_info["dist_range"]
    delta_dist = green_info["delta_dist"]
    time_reduction = green_info["time_reduction"]
    grn_dep_list = green_info["event_depth_list"]
    grn_receiver_list = green_info["receiver_depth_list"]
    if not isinstance(grn_dep_list, list):
        grn_dep_source = grn_dep_list
    else:
        grn_dep_source = grn_dep_list[
            np.argmin(np.abs(event_depth_km - np.array(grn_dep_list)))
        ]
    if not isinstance(grn_receiver_list, list):
        grn_dep_receiver = grn_receiver_list
    else:
        grn_dep_receiver = grn_receiver_list[
            np.argmin(np.abs(receiver_depth_km - np.array(grn_receiver_list)))
        ]

    # Common: calculate rotation matrix and convert moment tensor
    gamma = np.deg2rad(az_deg)
    A_rotate = create_rotate_z_mat(gamma=gamma)
    focal_mechanism = check_convert_fm(focal_mechanism)
    mt_ned_full = mt2full_mt_matrix(mt=focal_mechanism, flag="ned")
    mt_rotate = A_rotate.T @ mt_ned_full @ A_rotate
    mt_list = np.array(
        [
            mt_rotate[0, 0],
            mt_rotate[0, 1],
            mt_rotate[0, 2],
            mt_rotate[1, 1],
            mt_rotate[1, 2],
            mt_rotate[2, 2],
        ]
    )
    mt_rtp = convert_mt_axis(mt_list, "ned2rtp")

    # Select processing branch based on output_type
    if output_type in one_com_list:
        # One-component branch (e.g., gravimeter)
        u = seek_raw_qssp2020(
            path_green,
            grn_dep_source,
            grn_dep_receiver,
            dist_km,
            output_type,
            green_info,
        )
        u_enz_green_north = np.zeros((sampling_num, 1))
        for i_rtp in range(6):
            u_enz_green_north[:, 0] += mt_rtp[i_rtp] * u[:, i_rtp]
        seismograms = u_enz_green_north.T

    elif output_type in three_com_list:
        # Three-component branch: process data (e.g., displacement, velocity, acceleration)
        u_all = seek_raw_qssp2020(
            path_green,
            grn_dep_source,
            grn_dep_receiver,
            dist_km,
            output_type,
            green_info,
        )
        u_enz_green_north = np.zeros((sampling_num, 3))
        for i_rtp in range(6):
            for i_enz in range(3):
                u_enz_green_north[:, i_enz] += (
                    mt_rtp[i_rtp] * u_all[:, i_enz + 3 * i_rtp]
                )
        # Convert from green_north to RTZ coordinates
        u_rtz = np.zeros((sampling_num, 3))
        u_rtz[:, 0] = u_enz_green_north[:, 1]
        u_rtz[:, 1] = -u_enz_green_north[:, 0]
        u_rtz[:, 2] = u_enz_green_north[:, 2]
        seismograms = u_rtz.T

        if rotate:
            seismograms = rotate_rtz_to_enz(
                az_deg, r=u_rtz[:, 0], t=u_rtz[:, 1], z=u_rtz[:, 2]
            )

    elif output_type in six_com_list:
        # Six-component branch: process data (e.g., strain, strain rate, stress, stress rate)
        epsilon_all = seek_raw_qssp2020(
            path_green,
            grn_dep_source,
            grn_dep_receiver,
            dist_km,
            output_type,
            green_info,
        )
        epsilon_enz_green_north = np.tensordot(
            epsilon_all.reshape(sampling_num, 6, 6), mt_rtp, axes=(1, 0)
        )

        if rotate:
            seismograms = rotate_symmetric_tensor_series(
                epsilon_enz_green_north, gamma
            ).T
        else:
            seismograms = epsilon_enz_green_north.T
    else:
        raise ValueError(
            "output_type must be one of: disp, velo, acce, rota, rota_rate, gravimeter, gravitation, strain, strain_rate, stress, stress_rate"
        )

    # Compute the Green's function distance closest to the input distance.
    nearest_indice = round((dist_km - dist_range[0]) / delta_dist)
    grn_dist = dist_range[0] + nearest_indice * delta_dist

    # Process P-wave arrival times if before_p, shift, or pad_zeros is set.
    tpts_table = None
    if (before_p is not None) or shift or pad_zeros:
        grn_first_p, grn_first_s = read_tpts_table_qssp2020(
            path_green=path_green,
            event_depth_km=grn_dep_source,
            receiver_depth_km=grn_dep_receiver,
            ind=nearest_indice,
        )
        tpts_table = {"p_onset": grn_first_p, "s_onset": grn_first_s}

    ts_count = 0
    if before_p is not None:
        ts_count = round(
            (tpts_table["p_onset"] - time_reduction - before_p) * srate_grn
        )
    if pad_zeros and time_reduction != 0:
        if before_p is not None:
            raise ValueError("can not set before_p and pad_zeros together")
        ts_count = round(-time_reduction * srate_grn)
        before_p = tpts_table["p_onset"]
    seismograms = np.roll(seismograms, -ts_count)
    if ts_count > 0:
        seismograms[:, -ts_count:] = 0
    elif ts_count < 0:
        seismograms[:, :-ts_count] = 0

    # Shift the seismograms if required.
    first_p = None
    first_s = None
    if shift:
        seismograms, first_p, first_s = shift_green2real_tpts(
            seismograms=seismograms,
            tpts_table=tpts_table,
            srate=srate_grn,
            green_before_p=before_p,
            event_depth_km=event_depth_km,
            dist_in_km=dist_km,
            receiver_depth_km=receiver_depth_km,
            model_name=model_name,
        )

    # conv_shift = round(green_info["source_duration"] * srate_grn / 2)
    # if conv_shift != 0:
    #     seismograms = np.roll(seismograms, -conv_shift)
    #     seismograms[:, -conv_shift:] = 0

    seismograms_resample = np.zeros(
        (seismograms.shape[0], round(sampling_num * srate / srate_grn))
    )
    for i in range(seismograms.shape[0]):
        seismograms_resample[i] = resample(
            seismograms[i], srate_old=srate_grn, srate_new=srate, zero_phase=True
        )

    # Return results.
    if only_seismograms:
        return seismograms_resample
    else:
        return (
            seismograms_resample,
            tpts_table,
            first_p,
            first_s,
            grn_dep_source,
            grn_dep_receiver,
            grn_dist,
        )


if __name__ == "__main__":
    pass



import os
import struct

import numpy as np


def read_tpts_table(path_greenfunc: str, dist_in_km: float, green_info: dict) -> dict:
    """

    :param path_greenfunc:
    :param dist_in_km:
    :param green_info:
    :return:  {
        "p_onset": tp_onset,
        "p_takeoff": tp_takeoff,
        "p_slowness": tp_slowness,
        "s_onset": ts_onset,
        "s_takeoff": ts_takeoff,
        "s_slowness": ts_slowness
    }
    """
    dist_green_num = np.argmin(np.abs(np.array(green_info["dist_list"]) - dist_in_km))
    # print(dist_green_num)
    # print(green_info['dist_list'][dist_green_num])
    start_count = 4 + dist_green_num * 12

    def seek_time_table(start_count_in, phase):
        if phase == "P":
            fname = "tptable.dat"
        elif phase == "S":
            fname = "tstable.dat"
        else:
            raise ValueError("phase must be P or S")
        fr = open(os.path.join(path_greenfunc, fname), "rb")
        fr.seek(start_count_in)
        t_onset = struct.unpack("f", fr.read(4))[0]
        t_takeoff = struct.unpack("f", fr.read(4))[0]
        t_slowness = struct.unpack("f", fr.read(4))[0]
        fr.close()
        return [t_onset, t_takeoff, t_slowness]

    [tp_onset, tp_takeoff, tp_slowness] = seek_time_table(start_count, "P")
    [ts_onset, ts_takeoff, ts_slowness] = seek_time_table(start_count, "S")

    return {
        "p_onset": tp_onset,
        "p_takeoff": tp_takeoff,
        "p_slowness": tp_slowness,
        "s_onset": ts_onset,
        "s_takeoff": ts_takeoff,
        "s_slowness": ts_slowness,
    }



import os
import json

import numpy as np

from .read_tpts_table import read_tpts_table
from .utils import shift_green2real_tpts
from .focal_mechanism import check_convert_fm
from .geo import rotate_rtz_to_enz
from .signal_process import resample


def synthesize_spgrn2020(az_in_deg, time_series, focal_mechanism):
    """
    Kanamori, H., & Rivera, L. (2008).
    Source inversion of W phase: speeding up seismic tsunami warning.
    Geophysical Journal International, 175(1), 222–238.
    https://doi.org/10.1111/j.1365-246X.2008.03887.x
    """
    [M11, M12, M13, M22, M23, M33] = check_convert_fm(focal_mechanism=focal_mechanism)
    # expl.,strike-slip,dip-slip,clvd
    exp = (M11 + M22 + M33) / 3
    ss1 = -M12
    ss2 = (M11 - M22) / 2
    ds1 = M13
    ds2 = -M23
    clvd = M33 - exp

    az = np.deg2rad(180 - az_in_deg)
    # az = np.deg2rad(az_in_deg)
    sin_az, cos_az = np.sin(az), np.cos(az)
    sin_2az, cos_2az = np.sin(2 * az), np.cos(2 * az)
    m1 = [exp, ss1 * sin_2az + ss2 * cos_2az, ds1 * cos_az + ds2 * sin_az, clvd]
    m2 = [ss1 * cos_2az - ss2 * sin_2az, ds1 * sin_az - ds2 * cos_az]
    z = (
        time_series[0] * m1[0]
        + time_series[2] * m1[1]
        + time_series[5] * m1[2]
        + time_series[8] * m1[3]
    )
    r = (
        time_series[1] * m1[0]
        + time_series[3] * m1[1]
        + time_series[6] * m1[2]
        + time_series[9] * m1[3]
    )
    t = time_series[4] * m2[0] + time_series[7] * m2[1]

    seismograms = np.array([r, t, z])
    return seismograms


def read_time_series_spgrn2020(path_grn_data, dist_in_km, green_info):
    # dist_green, dist_green_num = find_nearest_dichotomy(
    #     dist_in_km, green_info["dist_list"]
    # )
    dist_green_num = np.argmin(np.abs(np.array(green_info["dist_list"]) - dist_in_km))
    dist_green = green_info["dist_list"][dist_green_num]
    N_T = round(green_info["time_window"] / green_info["sampling_interval"]) + 1
    length_each = 3 + (2 + N_T) * 10
    start_count = dist_green_num * length_each

    fr = open(path_grn_data, "rb")
    time_series = np.fromfile(
        file=fr, dtype=np.float32, count=length_each, offset=start_count * 4
    )
    time_series = time_series[3:].reshape(10, 2 + N_T)
    time_series = time_series[:, 1:-1]
    fr.close()
    # for i in range(10):
    #     plt.figure()
    #     plt.plot(time_series[i][:500])
    #     plt.show()
    return time_series, dist_green


def seek_spgrn2020(
    path_green,
    event_depth_km,
    receiver_depth_km,
    az_deg,
    dist_km,
    focal_mechanism,
    srate,
    rotate=True,
    before_p=20,
    pad_zeros=False,
    shift=False,
    only_seismograms=True,
    model_name="ak135",
    green_info=None,
):
    """

    :param path_green:
    :param event_depth_km:
    :param receiver_depth_km:
    :param az_deg:
    :param dist_km:
    :param focal_mechanism:
    :param srate:
    :param rotate:
    :param before_p:
    :param pad_zeros:
    :param shift:
    :param only_seismograms: only return seismograms
    :param model_name: earth model name in obspy TaupModel
    :return:
    """
    if green_info is None:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            green_info = json.load(fr)
    srate_grn = 1 / green_info["sampling_interval"]
    sampling_num = (
        round(green_info["time_window"] / green_info["sampling_interval"]) + 1
    )
    green_before_p = green_info["green_before_p"]
    grn_dep_list = green_info["event_depth_list"]
    grn_receiver_list = green_info["receiver_depth_list"]
    if not isinstance(grn_dep_list, list):
        grn_dep_source = grn_dep_list
    else:
        grn_dep_source = grn_dep_list[
            np.argmin(np.abs(event_depth_km - np.array(grn_dep_list)))
        ]
    if not isinstance(grn_receiver_list, list):
        grn_dep_receiver = grn_receiver_list
    else:
        grn_dep_receiver = grn_receiver_list[
            np.argmin(np.abs(receiver_depth_km - np.array(grn_receiver_list)))
        ]

    path_greenfunc = str(
        os.path.join(
            path_green, "GreenFunc", "%.2f" % grn_dep_source, "%.2f" % grn_dep_receiver
        )
    )

    # green_info = read_green_info_spgrn2020(
    #     path_greenfunc=path_greenfunc, green_depth=grn_dep_source
    # )
    tpts_table = read_tpts_table(
        path_greenfunc=path_greenfunc, dist_in_km=dist_km, green_info=green_info
    )

    path_grn_data = os.path.join(path_greenfunc, "grn_d%.2f" % grn_dep_source)
    time_series, grn_dist = read_time_series_spgrn2020(
        path_grn_data=path_grn_data, dist_in_km=dist_km, green_info=green_info
    )
    # r,t,z
    seismograms = synthesize_spgrn2020(
        az_in_deg=az_deg, time_series=time_series, focal_mechanism=focal_mechanism
    )
    if rotate:
        seismograms = rotate_rtz_to_enz(
            az_in_deg=az_deg, r=seismograms[0], t=seismograms[1], z=seismograms[2]
        )[:]

    ts_count = 0
    if before_p is not None:
        ts_count = round((green_before_p - before_p) * srate_grn)
    if pad_zeros:
        if before_p is not None:
            raise ValueError("can not set before_p and pad_zeros together")
        ts_count = round((green_before_p - tpts_table["p_onset"]) * srate_grn)
        before_p = tpts_table["p_onset"]
    seismograms = np.roll(seismograms, -ts_count)
    if ts_count > 0:
        seismograms[:, -ts_count:] = 0
    elif ts_count < 0:
        seismograms[:, :-ts_count] = 0

    first_p = None
    first_s = None
    if shift:
        seismograms, first_p, first_s = shift_green2real_tpts(
            seismograms=seismograms,
            tpts_table=tpts_table,
            srate=srate_grn,
            green_before_p=before_p,
            event_depth_km=event_depth_km,
            dist_in_km=dist_km,
            model_name=model_name,
        )

    # conv_shift = round(green_info["source_duration"] * srate_grn / 2)
    # if conv_shift != 0:
    #     seismograms = np.roll(seismograms, -conv_shift)
    #     seismograms[:, -conv_shift:] = 0

    seismograms_resasmple = np.zeros((3, round(sampling_num * srate / srate_grn)))
    for i in range(3):
        seismograms_resasmple[i] = resample(
            seismograms[i], srate_old=srate_grn, srate_new=srate, zero_phase=True
        )

    if only_seismograms:
        return seismograms_resasmple
    else:
        return (
            seismograms_resasmple,
            tpts_table,
            first_p,
            first_s,
            grn_dep_source,
            grn_dep_receiver,
            grn_dist,
        )


if __name__ == "__main__":
    pass



import os
import shutil
import platform
import pickle
import json
import datetime

from tqdm import tqdm
from multiprocessing import Pool
try:
    from mpi4py import MPI
except:
    pass

from .create_edgrn import create_inp_edgrn2, call_edgrn2
from .utils import group, convert_earth_model_nd2nd_without_Q


def _call_edgrn2_star(args):
    return call_edgrn2(*args)


def pre_process_edgrn2(
    processes_num,
    path_green,
    path_bin,
    grn_source_depth_range,
    grn_source_delta_depth,
    grn_dist_range,
    grn_delta_dist,
    obs_depth_list,
    wavenumber_sampling_rate=12,
    path_nd=None,
    earth_model_layer_num=None,
):
    # print("preprocessing edgrn2")
    if platform.system() == "Windows":
        path_bin_call = os.path.join(path_green, "edgrn2.exe")
    else:
        path_bin_call = os.path.join(path_green, "edgrn2.bin")
    shutil.copy(path_bin, path_bin_call)

    for obs_depth in obs_depth_list:
        sub_sub_dir = str(os.path.join(path_green, "edgrn2", "%.2f" % obs_depth))
        os.makedirs(sub_sub_dir, exist_ok=True)
        create_inp_edgrn2(
            path_green,
            obs_depth,
            grn_dist_range,
            grn_delta_dist,
            grn_source_depth_range,
            grn_source_delta_depth,
            wavenumber_sampling_rate,
            path_nd,
            earth_model_layer_num,
        )

    path_nd_without_Q = os.path.join(path_green, "noQ.nd")
    convert_earth_model_nd2nd_without_Q(path_nd, path_nd_without_Q)

    green_info = {
        "processes_num": processes_num,
        "grn_source_depth_range": grn_source_depth_range,
        "grn_source_delta_depth": grn_source_delta_depth,
        "grn_dist_range": grn_dist_range,
        "grn_delta_dist": grn_delta_dist,
        "obs_depth_list": obs_depth_list,
        "wavenumber_sampling_rate": wavenumber_sampling_rate,
        "path_nd": path_nd,
        "path_nd_without_Q": path_nd_without_Q,
        "earth_model_layer_num": earth_model_layer_num,
    }
    json_str = json.dumps(green_info, indent=4, ensure_ascii=False)
    with open(
        os.path.join(path_green, "green_lib_info.json"), "w", encoding="utf-8"
    ) as file:
        file.write(json_str)

    group_list_edgrn = group(obs_depth_list, processes_num)
    with open(os.path.join(path_green, "group_list_edgrn.pkl"), "wb") as fw:
        pickle.dump(group_list_edgrn, fw)  # type: ignore
    return group_list_edgrn


def create_grnlib_edgrn2_sequential(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list_edgrn.pkl"), "rb") as fr:
        group_list_edgrn = pickle.load(fr)
    for item in tqdm(group_list_edgrn, desc="Computing Green's function library"):
        for i in range(len(item)):
            # print("computing " + str(item[i]) + " km")
            call_edgrn2(item[i], path_green, check_finished)
    e = datetime.datetime.now()
    print("run time:%s" % str(e - s))
    return e - s


def create_grnlib_edgrn2_parallel(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list_edgrn.pkl"), "rb") as fr:
        group_list_edgrn = pickle.load(fr)
    tasks = []
    for grp in group_list_edgrn:
        for d in grp:
            tasks.append((d, path_green, check_finished))

    processes = None
    try:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            processes = json.load(fr).get("processes_num", None)
    except Exception:
        pass

    with Pool(processes=processes) as pool:
        for _ in tqdm(
            pool.imap_unordered(_call_edgrn2_star, tasks, chunksize=1),
            total=len(tasks),
            desc="Computing Green's function library",
        ):
            pass
    e = datetime.datetime.now()
    return e - s


def create_grnlib_edgrn2_parallel_multi_nodes(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list_edgrn.pkl"), "rb") as fr:
        group_list_edgrn = pickle.load(fr)
    for ind_group in range(len(group_list_edgrn)):
        comm = MPI.COMM_WORLD
        processes_num = comm.Get_size()
        rank = comm.Get_rank()
        if processes_num != len(group_list_edgrn[0]):
            raise ValueError(
                "processes_num is %d, item num in group is %d. \n"
                "Pleasse check the process num!"
                % (processes_num, len(group_list_edgrn[0]))
            )
        print("ind_group:%d rank:%d" % (ind_group, rank))
        call_edgrn2(
            obs_depth=group_list_edgrn[ind_group][rank],
            path_green=path_green,
            check_finished=check_finished,
        )
    e = datetime.datetime.now()
    print("run time:" + str(e - s))
    return e - s



from typing import Tuple
import warnings

import numpy as np


def convert_mt_axis(mt, convert_flag):
    """
    convert moment tensor from one axis to another axis.
    :param mt: moment tensor , if in ned axis, [M11, M12, M13, M22, M23, M33],
                               if in rtp axis, [Mrr, Mtt, Mpp, Mrt, Mrp, Mtp].
    :param convert_flag: 'ned2rtp' or 'rtp2ned'.
    :return:
    """
    if convert_flag == "ned2rtp":
        Mtt = mt[0]
        Mtp = -mt[1]
        Mrt = mt[2]
        Mpp = mt[3]
        Mrp = -mt[4]
        Mrr = mt[5]
        mt = [Mrr, Mtt, Mpp, Mrt, Mrp, Mtp]
    elif convert_flag == "rtp2ned":
        M11 = mt[1]
        M12 = -mt[5]
        M13 = mt[3]
        M22 = mt[2]
        M23 = -mt[4]
        M33 = mt[0]
        mt = [M11, M12, M13, M22, M23, M33]
    return mt


def mt2full_mt_matrix(mt, flag="ned") -> np.ndarray:
    """
    create full moment tensor matrix from 6 components.
    :param mt: in NED axis, [M11, M12, M13, M22, M23, M33].
               in rtp axis, [mrr, mtt, mpp, mrt, mrp, mtp]
    :param flag: 'ned'/'rtp'
    :return: full moment tensor matrix,
    in NED axis,
    np.array([[M11, M12, M13],
              [M12, M22, M23],
              [M13, M23, M33]])
    in rtp axis,
    np.array([[mrr, mrt, mrp],
              [mrt, mtt, mtp],
              [mrp, mtp, mpp],
            ])
    """
    mpq = np.zeros((3, 3))
    if flag == "ned":
        mpq[0, 0] = mt[0]
        mpq[0, 1] = mt[1]
        mpq[0, 2] = mt[2]
        mpq[1, 0] = mpq[0, 1]
        mpq[1, 1] = mt[3]
        mpq[1, 2] = mt[4]
        mpq[2, 0] = mpq[0, 2]
        mpq[2, 1] = mpq[1, 2]
        mpq[2, 2] = mt[5]
    elif flag == "rtp":
        mpq[0, 0] = mt[0]
        mpq[0, 1] = mt[3]
        mpq[0, 2] = mt[4]
        mpq[1, 0] = mpq[0, 1]
        mpq[1, 1] = mt[1]
        mpq[1, 2] = mt[5]
        mpq[2, 0] = mpq[0, 2]
        mpq[2, 1] = mpq[1, 2]
        mpq[2, 2] = mt[2]
    else:
        raise ValueError("axis flag wrong")
    return mpq


def moment_from_moment_tensor(mt):
    m0 = np.sqrt(
        1
        / 2
        * (
            mt[0] ** 2
            + 2 * mt[1] ** 2
            + 2 * mt[2] ** 2
            + mt[3] ** 2
            + 2 * mt[4] ** 2
            + mt[5] ** 2
        )
    )
    return m0


def check_convert_fm(focal_mechanism):
    """

    :param focal_mechanism:
    :return: [M11, M12, M13, M22, M23, M33]
    """
    if len(focal_mechanism) == 3:
        mt = plane2mt(1, focal_mechanism[0], focal_mechanism[1], focal_mechanism[2])
        [M11, M12, M13, M22, M23, M33] = list(mt)
    elif len(focal_mechanism) == 4:
        mt = plane2mt(
            focal_mechanism[0],
            focal_mechanism[1],
            focal_mechanism[2],
            focal_mechanism[3],
        )
        [M11, M12, M13, M22, M23, M33] = list(mt)
    elif len(focal_mechanism) == 6:
        [M11, M12, M13, M22, M23, M33] = focal_mechanism
    elif len(focal_mechanism) == 7:
        M0 = focal_mechanism[0]
        temp = np.array(focal_mechanism[1:])
        M0_temp = moment_from_moment_tensor(temp)
        temp = temp / M0_temp
        M11 = M0 * temp[0]
        M12 = M0 * temp[1]
        M13 = M0 * temp[2]
        M22 = M0 * temp[3]
        M23 = M0 * temp[4]
        M33 = M0 * temp[5]
    else:
        raise ValueError("focal mechanism wrong")
    return [M11, M12, M13, M22, M23, M33]


def cal_m0_from_mt(mt):
    m0 = np.sqrt(
        1
        / 2
        * (
            mt[0] ** 2
            + 2 * mt[1] ** 2
            + 2 * mt[2] ** 2
            + mt[3] ** 2
            + 2 * mt[4] ** 2
            + mt[5] ** 2
        )
    )
    return m0


def mt2plane(mt):
    """

    :param mt: in NED axis, [M11, M12, M13, M22, M23, M33].
    :return: [[strike1, dip1, rake1], [strike2, dip2, rake2],
    n1, d1, n2, d2, t, b, p, eigenvalues]

    n is the normal vector of the plane, in NED axis.
    d is the rupture vector on the plane, in NED axis.

    n points in the negative direction of D, i.e. upwards

    When the dip angle is 0,

    return: [pl1, pl2, n1, d1, n2, d2, t, b, p, eigenvalues]
    """
    M = mt2full_mt_matrix(mt)
    [eigenvalues, eigenvectors] = np.linalg.eig(M)

    index = eigenvalues.argsort()
    eigenvectors = eigenvectors[:, index]
    eigenvalues = eigenvalues[index]
    p = eigenvectors[:, 0]
    b = eigenvectors[:, 1]
    t = eigenvectors[:, 2]
    n = 1 / np.sqrt(2) * (t + p)
    d = 1 / np.sqrt(2) * (t - p)

    threshold = 1e-15

    def ignore_small_angle_vector(vector):
        for i in range(3):
            if np.abs(vector[i]) < threshold:
                vector[i] = 0
            if np.abs(vector[i] - 1) < threshold:
                vector[i] = 1
            if np.abs(vector[i] + 1) < threshold:
                vector[i] = -1
        return vector

    n = ignore_small_angle_vector(n)
    d = ignore_small_angle_vector(d)

    # print(eigenvalues)
    # print(eigenvectors)
    # print(t, b, p)
    # print(n,d)

    def nd2plane(n_in, d_in):
        # 保证n朝上
        if n_in[2] > 0:
            n_in = -n_in
            d_in = -d_in
        delta = np.arccos(-n_in[2])
        if np.abs(delta - 0) <= threshold:
            delta = 0
        if np.abs(delta - np.pi / 2) <= threshold:
            delta = np.pi / 2

        if n_in[1] == 0:
            if delta == 0:
                warnings.warn("n is vertical. Strike is set as 0.")
                phi = 0
            elif delta == np.pi / 2:
                warnings.warn(
                    "n is horizontal. The part directed by n is set as the hanging wall of the fault."
                )
                if n_in[0] > 0:
                    phi = np.pi * 3 / 2
                else:
                    phi = np.pi / 2
            else:
                if n_in[0] > 0:
                    phi = np.pi * 3 / 2
                else:
                    phi = np.pi / 2
        else:
            if delta == np.pi / 2:
                warnings.warn(
                    "n is horizontal. The part directed by n is set as the hanging wall of the fault."
                )
                if n_in[0] == 0:
                    if n_in[1] == 1:
                        phi = 0
                    elif n_in[1] == -1:
                        phi = np.pi
                    else:
                        raise ValueError(
                            "n is horizontal,n[0] is 0, but n[1] is not 1 or -1."
                        )
                else:
                    phi = np.arctan(-n_in[0] / n_in[1])
            else:
                phi = np.arctan(-n_in[0] / n_in[1])
        if (n_in[0] <= 0) and (n_in[1] > 0):
            pass
        elif (n_in[0] <= 0) and (n_in[1] < 0):
            phi = phi + np.pi
        elif (n_in[0] > 0) and (n_in[1] < 0):
            phi = phi + np.pi
        elif (n_in[0] > 0) and (n_in[1] > 0):
            phi = phi + 2 * np.pi

        cos_lambda = d_in[0] * np.cos(phi) + d_in[1] * np.sin(phi)
        sin_lambda_cos_delta = d_in[0] * np.sin(phi) - d_in[1] * np.cos(phi)
        lambda_ = np.arccos(cos_lambda)
        if sin_lambda_cos_delta < 0:
            lambda_ = -lambda_
        # elif np.abs(sin_lambda_cos_delta) <= threshold:
        #     if d_in[0] != 0:
        #         lambda_ = np.arctan(d_in[1] / d_in[0])
        #     else:
        #         if np.abs(d_in[1] - 1) <= threshold:
        #             lambda_ = 0
        #         else:
        #             lambda_ = np.arccos(-d_in[1])

        pl = np.array([phi, delta, lambda_])
        pl = pl * 180 / np.pi
        return pl, n_in, d_in

    pl1, n1, d1 = list(nd2plane(n, d))
    pl2, n2, d2 = list(nd2plane(d, n))

    return [pl1, pl2, n1, d1, n2, d2, t, b, p, eigenvalues]


def plane2mt(M0, strike, dip, rake):
    """

    :param M0: scalar moment, unit: Nm
    :param strike: strike angle, unit: degree
    :param dip: dip angle, unit: degree
    :param rake: rake angle, unit: degree
    :return: mt : numpy array
        in NEZ(NED) axis, [M11, M12, M13, M22, M23, M33].
    """
    strike, dip, rake = strike * np.pi / 180, dip * np.pi / 180, rake * np.pi / 180

    sin_strike, cos_strike = np.sin(strike), np.cos(strike)
    sin_2strike, cos_2strike = np.sin(2 * strike), np.cos(2 * strike)
    sin_dip, cos_dip = np.sin(dip), np.cos(dip)
    sin_2dip, cos_2dip = np.sin(2 * dip), np.cos(2 * dip)
    sin_lambda, cos_lambda = np.sin(rake), np.cos(rake)

    M11 = -M0 * (
        sin_dip * cos_lambda * sin_2strike + sin_2dip * sin_lambda * sin_strike**2
    )  # Mtt
    M12 = M0 * (
        sin_dip * cos_lambda * cos_2strike + 1 / 2 * sin_2dip * sin_lambda * sin_2strike
    )  # -Mtp
    M13 = -M0 * (
        cos_dip * cos_lambda * cos_strike + cos_2dip * sin_lambda * sin_strike
    )  # Mrt
    M22 = M0 * (
        sin_dip * cos_lambda * sin_2strike - sin_2dip * sin_lambda * cos_strike**2
    )  # Mpp
    M23 = -M0 * (
        cos_dip * cos_lambda * sin_strike - cos_2dip * sin_lambda * cos_strike
    )  # -Mrp
    M33 = M0 * sin_2dip * sin_lambda  # Mrr

    mt = np.array([M11, M12, M13, M22, M23, M33])
    return mt


def plane2nd(strike, dip, rake) -> Tuple[np.ndarray, np.ndarray]:
    """

    :param strike: unit: degree
    :param dip: unit: degree
    :param rake: unit: degree
    :return: n, np.ndarray
             normal vector of the fault plane, in NED axis.
             d, np.ndarray
             rupture vector on the fault plane, in NED axis.
    """
    strike, dip, rake = strike * np.pi / 180, dip * np.pi / 180, rake * np.pi / 180
    sin_strike, cos_strike = np.sin(strike), np.cos(strike)
    sin_dip, cos_dip = np.sin(dip), np.cos(dip)
    sin_rake, cos_rake = np.sin(rake), np.cos(rake)

    n_nwu = np.array([-sin_dip * sin_strike, -sin_dip * cos_strike, cos_dip])
    n = np.array([n_nwu[0], -n_nwu[1], -n_nwu[2]])

    d_nwu = np.array(
        [
            cos_rake * cos_strike + sin_rake * cos_dip * sin_strike,
            -cos_rake * sin_strike + sin_rake * cos_dip * cos_strike,
            sin_rake * sin_dip,
        ]
    )
    d = np.array([d_nwu[0], -d_nwu[1], -d_nwu[2]])
    if n[2] > 0:  # 保证n朝上
        n = -n
        d = -d
    return n, d


def plane2tbp(strike, dip, rake) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    :param strike: unit: degree
    :param dip: unit: degree
    :param rake: unit: degree
    :return: [np.array(t), np.array(b), np.array(p)]
    """
    n, d = plane2nd(strike, dip, rake)
    t = 1 / np.sqrt(2) * (n + d)
    p = 1 / np.sqrt(2) * (n - d)
    b = np.cross(t, p)
    if t[2] < 0:
        t = -t
    if b[2] < 0:
        b = -b
    if p[2] < 0:
        p = -p
    return t, b, p



s="""# This is the input file of FORTRAN77 program "qseis06" for calculation of
# synthetic seismograms based on a layered halfspace earth model.
#
# by
# Rongjiang  Wang <wang@gfz-potsdam.de>
# GeoForschungsZentrum Potsdam
# Telegrafenberg, D-14473 Potsdam, Germany
#
# Last modified: Potsdam, Nov., 2006
#
# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
# If not specified, SI Unit System is used overall!
#
# Coordinate systems:
# cylindrical (z,r,t) with z = downward,
#                          r = from source outward,
#                          t = azmuth angle from north to east;
# cartesian (x,y,z) with   x = north,
#                          y = east,
#                          z = downward;
# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
#
#	SOURCE PARAMETERS
#	=================
# 1. source depth [km]
#------------------------------------------------------------------------------
1.00
#------------------------------------------------------------------------------
#
#	RECEIVER PARAMETERS
#	===================
# 1. receiver depth [km]
# 2. switch for distance sampling role (1/0 = equidistant/irregular); switch
#    for unit used (1/0 = km/deg)
# 3. number of distance samples
# 4. if equidistant, then start and end trace distance (> 0); else distance
#    list (please order the receiver distances from small to large)
# 5. (reduced) time begin [sec] & length of time window [sec], number of time
#    samples (<= 2*nfmax in qsglobal.h)
# 6. switch for unit of the following time reduction parameter: 1 = velocity
#    [km/sec], 0 = slowness [sec/deg]; time reduction parameter
#------------------------------------------------------------------------------
0.00
1 1
200
2.000000 400.000000
0.000000 255.500000 512
1 0.000000
#------------------------------------------------------------------------------
#
#	WAVENUMBER INTEGRATION PARAMETERS
#	=================================
# 1. select slowness integration algorithm (0 = suggested for full wave-field
#    modelling; 1 or 2 = suggested when using a slowness window with narrow
#    taper range - a technique for suppressing space-domain aliasing);
# 2. 4 parameters for low and high slowness (Note 1) cut-offs [s/km] with
#    tapering: 0 < slw1 < slw2 defining cosine taper at the lower end, and 0 <
#    slw3 < slw4 defining the cosine taper at the higher end. default values
#    will be used in case of inconsistent input of the cut-offs (possibly with
#    much more computational effort);
# 3. parameter for sampling rate of the wavenumber integration (1 = sampled
#    with the spatial Nyquist frequency, 2 = sampled with twice higher than
#    the Nyquist, and so on: the larger this parameter, the smaller the space-
#    domain aliasing effect, but also the more computation effort); 
# 4. the factor for suppressing time domain aliasing (> 0 and <= 1) (Note 2).
#------------------------------------------------------------------------------
0
0.000000 0.000000 0.000000 0.000000
12.000000
0.010000
#------------------------------------------------------------------------------
#
#	        OPTIONS FOR PARTIAL SOLUTIONS
#       (only applied to the source-site structure)
#	    ===========================================
#
# 1. switch for filtering free surface effects (0 = with free surface, i.e.,
#    do not select this filter; 1 = without free surface; 2 = without free
#    surface but with correction on amplitude and wave form. Note switch 2
#    can only be used for receivers at the surface)
# 2. switch for filtering waves with a shallow penetration depth (concerning
#    their whole trace from source to receiver), penetration depth limit [km]
#
#    if this option is selected, waves whose travel path never exceeds the
#    given depth limit will be filtered ("seismic nuting"). the condition for
#    selecting this filter is that the given shallow path depth limit should
#    be larger than both source and receiver depth.
#
# 3. number of depth ranges where the following selected up/down-sp2oing P or
#    SV waves should be filtered
# 4. the 1. depth range: upper and lower depth [km], switch for filtering P
#    or SV wave in this depth range:
#
#    switch no:              1      2        3       4         other
#    filtered phase:         P(up)  P(down)  SV(up)  SV(down)  Error
#
# 5. the 2. ...
#
#    The partial solution options are useful tools to increase the numerical
#    significance of desired wave phases. Especially when the desired phases
#    are smaller than the undesired phases, these options should be selected
#    and carefully combined.
#------------------------------------------------------------------------------
0
0   0.00
0
#0.0 0.0 2
#0.0 0.4 4
#------------------------------------------------------------------------------
#
#	SOURCE TIME FUNCTION (WAVELET) PARAMETERS (Note 3)
#	==================================================
# 1. wavelet duration [unit = time sample rather than sec!], that is about
#    equal to the half-amplitude cut-off period of the wavelet (> 0. if <= 0,
#    then default value = 2 time samples will be used), and switch for the
#    wavelet form (0 = user's own wavelet; 1 = default wavelet: normalized
#    square half-sinusoid for simulating a physical delta impulse; 2 = tapered
#    Heaviside wavelet, i.e. integral of wavelet 1)
# 2. IF user's own wavelet is selected, then number of the wavelet time samples
#    (<= 1024), and followed by
# 3. equidistant wavelet time samples
# 4  ...(continue) (! no comment lines allowed between the time sample list!)
#    IF default, delete line 2, 3, 4 ... or comment them out!
#------------------------------------------------------------------------------
5 2
#------------------------------------------------------------------------------
#
#	 FILTER PARAMETERS OF RECEIVERS (SEISMOMETERS OR HYDROPHONES)
#	 ============================================================
# 1. constant coefficient (normalization factor)
# 2. number of roots (<= nrootmax in qsglobal.h)
# 3. list of the root positions in the complex format (Re,Im). If no roots,
#    comment out this line
# 4. number of poles (<= npolemax in qsglobal.h)
# 5. list of the pole positions in the complex format (Re,Im). If no poles,
#    comment out this line
#------------------------------------------------------------------------------
(1.0, 0.0)
0
# (0.0, 0.0), (0.0, 0.0)
0
# (-4.35425, 4.44222), (-4.35425,-4.44222)
#------------------------------------------------------------------------------
#
#	OUTPUT FILES FOR GREEN'S FUNCTIONS (Note 4)
#	===========================================
# 1. selections of source types (yes/no = 1/0)
# 2. file names of Green's functions (please give the names without extensions,
#    which will be appended by the program automatically: *.tz, *.tr, *.tt
#    and *.tv are for the vertical, radial, tangential, and volume change (for
#    hydrophones) components, respectively)
#------------------------------------------------------------------------------
#  explosion   strike-slip  dip-slip  clvd  single_f_v  single_f_h
#------------------------------------------------------------------------------
1    1    1    1    0    0
'ex' 'ss' 'ds' 'cl' 'fz' 'fh'
#------------------------------------------------------------------------------
#	OUTPUT FILES FOR AN ARBITRARY POINT DISLOCATION SOURCE
#               (for applications to earthquakes)
#	======================================================
# 1. selection (0 = not selected; 1 or 2 = selected), if (selection = 1), then
#    the 6 moment tensor elements [N*m]: Mxx, Myy, Mzz, Mxy, Myz, Mzx (x is
#    northward, y is eastward and z is downard); else if (selection = 2), then
#    Mis [N*m] = isotropic moment part = (MT+MN+MP)/3, Mcl = CLVD moment part
#    = (2/3)(MT+MP-2*MN), Mdc = double-couple moment part = MT-MN, Strike [deg],
#    Dip [deg] and Rake [deg].
#
#    Note: to use this option, the Green's functions above should be computed
#          (selection = 1) if they do not exist already.
# 2. switch for azimuth distribution of the stations (0 = uniform azimuth,
#    else = irregular azimuth angles)
# 3. list of the azimuth angles [deg] for all stations given above (if the
#    uniform azimuth is selected, then only one azimuth angle is required)
#
#------------------------------------------------------------------------------
#     Mis        Mcl        Mdc        Strike     Dip        Rake      File
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#     Mxx        Myy        Mzz        Mxy        Myz        Mzx       File
#------------------------------------------------------------------------------
0
0
0
#------------------------------------------------------------------------------
#
#	GLOBAL MODEL PARAMETERS (Note 5)
#	================================
# 1. switch for flat-earth-transform
# 2. gradient resolution [%] of vp, vs, and ro (density), if <= 0, then default
#    values (depending on wave length at cut-off frequency) will be used
#------------------------------------------------------------------------------
0
0  0  0
#------------------------------------------------------------------------------
#
#	                LAYERED EARTH MODEL
#       (SHALLOW SOURCE + UNIFORM DEEP SOURCE/RECEIVER STRUCTURE)
#	=========================================================
# 1. number of data lines of the layered model (source site) 
#------------------------------------------------------------------------------
25
#------------------------------------------------------------------------------
#
#	MULTILAYERED MODEL PARAMETERS (source site)
#	===========================================
# no  depth[km]    vp[km/s]    vs[km/s]   ro[g/cm^3]   qp      qs
#------------------------------------------------------------------------------
1  0.00  5.9000  3.4400  2.6700  927.34  599.99
2  15.53  5.9000  3.4400  2.6700  927.34  599.99
3  15.53  6.3000  3.6200  2.7400  927.34  599.99
4  25.26  6.3000  3.6200  2.7400  927.34  599.99
5  25.26  6.9000  3.8700  2.9100  927.34  599.99
6  35.00  6.9000  3.8700  2.9100  927.34  599.99
7  35.00  8.0574  4.4817  3.3509  927.34  599.99
8  49.50  8.0574  4.4817  3.3509  927.34  599.99
9  49.50  8.0545  4.4836  3.3864  927.34  599.99
10  51.50  8.0545  4.4836  3.3864  927.34  599.99
11  51.50  8.0539  4.4841  3.3939  927.34  599.99
12  53.00  8.0539  4.4841  3.3939  927.34  599.99
13  53.00  8.0493  4.4871  3.4492  927.34  599.99
14  77.50  8.0493  4.4871  3.4492  114.87  75.60
15  77.50  8.0477  4.4949  3.4642  114.87  75.60
16  120.00  8.0477  4.4949  3.4642  115.30  76.06
17  120.00  8.0816  4.5022  3.4128  115.30  76.06
18  142.50  8.0816  4.5022  3.4128  115.30  76.06
19  142.50  8.1438  4.5067  3.3849  115.30  76.06
20  165.00  8.1438  4.5067  3.3849  118.37  76.55
21  165.00  8.2064  4.5113  3.3593  118.37  76.55
22  187.50  8.2064  4.5113  3.3593  118.37  76.55
23  187.50  8.2692  4.5160  3.3359  118.37  76.55
24  210.00  8.2692  4.5160  3.3359  125.22  79.40
25  210.00  8.3309  4.5335  3.3313  125.22  79.40
#------------------------------------------------------------------------------
#
#	          LAYERED EARTH MODEL
#       (ONLY THE SHALLOW RECEIVER STRUCTURE)
#       =====================================
# 1. number of data lines of the layered model
#
#    Note: if the number = 0, then the receiver site is the same as the
#          source site, else different receiver-site structure is considered.
#          please be sure that the lowest interface of the receiver-site
#          structure given given below can be found within the source-site
#          structure, too.
#
#------------------------------------------------------------------------------
0
#------------------------------------------------------------------------------
#
#	MULTILAYERED MODEL PARAMETERS (shallow receiver-site structure)
#	===============================================================
# no  depth[km]    vp[km/s]    vs[km/s]   ro[g/cm^3]   qp      qs
#------------------------------------------------------------------------------
#---------------------------------end of all inputs----------------------------"""


import os
import math

import numpy as np
import pandas as pd

from .qseis06inp import s as str_inp
from .utils import convert_earth_model_nd2inp, call_exe


def create_dir_qseis06(
    path_green,
    event_depth,
    receiver_depth,
    dist_range,
    delta_dist,
    N_each_group=100,
    order=0,
):
    sub_dir = str(
        os.path.join(path_green, "%.2f" % event_depth, "%.2f" % receiver_depth)
    )
    N_dist = math.ceil((dist_range[1] - dist_range[0]) / delta_dist) + 1
    N_dist_group = math.ceil(N_dist / N_each_group)
    for n in range(N_dist_group):
        if receiver_depth > 0:
            for o in range(2 * order + 1):
                os.makedirs(os.path.join(sub_dir, "%d_%d" % (n, o)), exist_ok=True)
        else:
            for o in range(order + 1):
                os.makedirs(os.path.join(sub_dir, "%d_%d" % (n, o)), exist_ok=True)
    return N_dist, N_dist_group


def create_inp_qseis06(
    path_sub_dir,
    event_depth,
    receiver_depth,
    dist_range,
    delta_dist,
    N_dist,
    N_dist_group,
    N_each_group,
    time_window,
    sampling_interval,
    slowness_int_algorithm=0,
    slowness_window=None,
    time_reduction_velo=0,
    wavenumber_sampling_rate=12,
    anti_alias=0.01,
    free_surface=True,
    wavelet_duration=4,
    wavelet_type=1,
    flat_earth_transform=True,
    path_nd=None,
    earth_model_layer_num=None,
    order=0,
):
    # when receiver depth is not 0, change dists in inp file
    r_ratio = (6371 - receiver_depth) / 6371
    lines = str_inp.split("\n")
    lines = [line + "\n" for line in lines]

    lines_earth = lines[207:-22]
    lines_end = lines[-22:]

    # SOURCE PARAMETERS
    lines[26] = "%.2f\n" % event_depth

    # RECEIVER PARAMETERS
    lines[42] = "%.2f\n" % receiver_depth
    lines[43] = "1 1\n"
    lines[46] = "%f %f %d\n" % (
        0.0,
        time_window,
        round(time_window / sampling_interval + 1),
    )
    lines[47] = "%d %f\n" % (1, time_reduction_velo)

    # WAVENUMBER INTEGRATION PARAMETERS
    lines[66] = "%d\n" % slowness_int_algorithm
    if slowness_window is not None:
        lines[67] = "%f %f %f %f\n" % (
            slowness_window[0],
            slowness_window[1],
            slowness_window[2],
            slowness_window[3],
        )
    else:
        lines[67] = "0.0 0.0 0.0 0.0\n"
    lines[68] = "%f\n" % wavenumber_sampling_rate
    lines[69] = "%f\n" % anti_alias

    # OPTIONS FOR PARTIAL SOLUTIONS
    if free_surface:
        lines[103] = "0\n"
    else:
        lines[103] = "1\n"

    # SOURCE TIME FUNCTION (WAVELET) PARAMETERS (Note 3)
    lines[124] = "%d %d\n" % (wavelet_duration, wavelet_type)

    # GLOBAL MODEL PARAMETERS (Note 5)
    if flat_earth_transform:
        lines[191] = "1\n"
    else:
        lines[191] = "0\n"

    if path_nd is not None:
        lines_earth = convert_earth_model_nd2inp(
            path_nd=path_nd, path_output="earth_model.dat"
        )
    if earth_model_layer_num is None:
        earth_model_layer_num = len(lines_earth)
    else:
        lines_earth = lines_earth[:earth_model_layer_num]
    lines[200] = "%d\n" % earth_model_layer_num
    lines = lines[:207] + lines_earth + lines_end

    for n in range(N_dist_group - 1):
        lines[44] = "%d\n" % N_each_group
        lines[45] = "%f %f\n" % (
            (dist_range[0] + n * N_each_group * delta_dist) * r_ratio,
            (dist_range[0] + ((n + 1) * N_each_group - 1) * delta_dist) * r_ratio,
        )
        path_inp = os.path.join(path_sub_dir, "%d_%d" % (n, order), "grn.inp")
        with open(path_inp, "w") as fw:
            fw.writelines(lines)
    else:
        res = N_dist - (N_dist_group - 1) * N_each_group
        lines[44] = "%d\n" % res
        lines[45] = "%f %f\n" % (
            (dist_range[0] + (N_dist_group - 1) * N_each_group * delta_dist) * r_ratio,
            dist_range[1] * r_ratio,
        )
        path_inp = os.path.join(
            path_sub_dir, "%d_%d" % (N_dist_group - 1, order), "grn.inp"
        )
        with open(path_inp, "w") as fw:
            fw.writelines(lines)
    return path_inp


def create_inp_qseis06_points(
    path_green,
    event_depth,
    receiver_depth,
    n_group,
    points,
    time_window,
    sampling_interval,
    slowness_int_algorithm=0,
    slowness_window=None,
    time_reduction_velo=0,
    wavenumber_sampling_rate=2,
    anti_alias=0.01,
    free_surface=True,
    wavelet_duration=4,
    wavelet_type=1,
    flat_earth_transform=True,
    path_nd=None,
    earth_model_layer_num=None,
    order=0,
):
    path_sub_dir = str(
        os.path.join(path_green, "%.2f" % event_depth, "%.2f" % receiver_depth)
    )
    # when receiver depth is not 0, change dists in inp file
    r_ratio = (6371 - receiver_depth) / 6371
    points = points * r_ratio

    lines = str_inp.split("\n")
    lines = [line + "\n" for line in lines]

    lines_earth = lines[207:-22]
    lines_end = lines[-22:]

    # SOURCE PARAMETERS
    lines[26] = "%f\n" % event_depth

    # RECEIVER PARAMETERS
    lines[42] = "%f\n" % receiver_depth
    lines[43] = "0 1\n"
    lines[44] = "%d\n" % len(points)
    lines[45] = " ".join("%.4f" % _ for _ in points) + "\n"
    lines[46] = "%f %f %d\n" % (
        0.0,
        time_window,
        round(time_window / sampling_interval + 1),
    )
    lines[47] = "%d %f\n" % (1, time_reduction_velo)

    # WAVENUMBER INTEGRATION PARAMETERS
    lines[66] = "%d\n" % slowness_int_algorithm
    if slowness_window is not None:
        lines[67] = "%f %f %f %f\n" % (
            slowness_window[0],
            slowness_window[1],
            slowness_window[2],
            slowness_window[3],
        )
    else:
        lines[67] = "0.0 0.0 0.0 0.0\n"
    lines[68] = "%f\n" % wavenumber_sampling_rate
    lines[69] = "%f\n" % anti_alias

    # OPTIONS FOR PARTIAL SOLUTIONS
    if free_surface:
        lines[103] = "0\n"
    else:
        lines[103] = "1\n"

    # SOURCE TIME FUNCTION (WAVELET) PARAMETERS (Note 3)
    lines[124] = "%d %d\n" % (wavelet_duration, wavelet_type)

    # GLOBAL MODEL PARAMETERS (Note 5)
    if flat_earth_transform:
        lines[191] = "1\n"
    else:
        lines[191] = "0\n"

    if path_nd is not None:
        lines_earth = convert_earth_model_nd2inp(
            path_nd=path_nd, path_output="earth_model.dat"
        )
    if earth_model_layer_num is None:
        earth_model_layer_num = len(lines_earth)
    else:
        lines_earth = lines_earth[:earth_model_layer_num]
    lines[200] = "%d\n" % earth_model_layer_num
    lines = lines[:207] + lines_earth + lines_end

    path_inp = os.path.join(path_sub_dir, "%d_%d" % (n_group, order), "grn.inp")
    with open(path_inp, "w") as fw:
        fw.writelines(lines)
    return path_inp


def call_qseis06(
    event_depth, receiver_depth, n_group, order, path_green, check_finished=False
):
    sub_sub_dir = str(
        os.path.join(
            path_green,
            "%.2f" % event_depth,
            "%.2f" % receiver_depth,
            "%d_%d" % (n_group, order),
        )
    )
    os.chdir(sub_sub_dir)
    path_inp = os.path.join(sub_sub_dir, "grn.inp")
    path_finished = os.path.join(sub_sub_dir, ".finished")

    if (
        check_finished
        and os.path.exists(path_finished)
        and len(os.listdir(sub_sub_dir)) > 2
    ):
        with open(path_finished, "r") as fr:
            output = fr.readlines()
        return output

    output = call_exe(
        path_green=path_green,
        path_inp=path_inp,
        path_finished=path_finished,
        name="qseis06",
    )
    return output


def convert_pd2bin_qseis06(path_greenfunc, remove=False):
    for com in ["tr", "tz", "tv"]:
        ex_com = pd.read_csv(
            str(os.path.join(path_greenfunc, "ex.%s" % com)), sep="\\s+"
        ).to_numpy()
        ss_com = pd.read_csv(
            str(os.path.join(path_greenfunc, "ss.%s" % com)), sep="\\s+"
        ).to_numpy()
        ds_com = pd.read_csv(
            str(os.path.join(path_greenfunc, "ds.%s" % com)), sep="\\s+"
        ).to_numpy()
        cl_com = pd.read_csv(
            str(os.path.join(path_greenfunc, "cl.%s" % com)), sep="\\s+"
        ).to_numpy()
        time_series_com = np.concatenate(
            [
                ex_com[:, 1:],
                ss_com[:, 1:],
                ds_com[:, 1:],
                cl_com[:, 1:],
            ]
        )
        time_series_com = np.array(time_series_com, dtype=np.float32)
        np.save(os.path.join(path_greenfunc, "grn_%s.npy" % com), time_series_com)
        if remove:
            os.remove(os.path.join(path_greenfunc, "ex.%s" % com))
            os.remove(os.path.join(path_greenfunc, "ss.%s" % com))
            os.remove(os.path.join(path_greenfunc, "ds.%s" % com))
            os.remove(os.path.join(path_greenfunc, "cl.%s" % com))

    for com in ["tt"]:
        ss_r = pd.read_csv(
            str(os.path.join(path_greenfunc, "ss.%s" % com)), sep="\\s+"
        ).to_numpy()
        ds_r = pd.read_csv(
            str(os.path.join(path_greenfunc, "ds.%s" % com)), sep="\\s+"
        ).to_numpy()
        time_series_com = np.concatenate(
            [
                ss_r[:, 1:],
                ds_r[:, 1:],
            ]
        )
        time_series_com = np.array(time_series_com, dtype=np.float32)
        np.save(os.path.join(path_greenfunc, "grn_%s.npy" % com), time_series_com)
        if remove:
            os.remove(os.path.join(path_greenfunc, "ss.%s" % com))
            os.remove(os.path.join(path_greenfunc, "ds.%s" % com))


if __name__ == "__main__":
    pass



import os
import json

import numpy as np
import pandas as pd
import scipy.signal as signal

from .focal_mechanism import check_convert_fm
from .geo import rotate_rtz_to_enz, rotate_symmetric_tensor_series
from .pytaup import read_tpts_table
from .utils import read_layerd_material, shift_green2real_tpts
from .signal_process import resample

# Define output type lists
one_com_list = ["volume"]
three_com_list = ["disp", "velo"]
rota_com_list = ["rota", "rota_rate"]
six_com_list = ["strain", "strain_rate", "stress", "stress_rate"]


def get_outfile_name_list(output_type):
    if output_type == "volume":
        name_list_psv = ["tv"]
        name_list_sh = []
    elif output_type == "disp" or output_type == "velo":
        name_list_psv = ["tz", "tr"]
        name_list_sh = ["tt"]
    elif output_type == "strain" or output_type == "strain_rate":
        name_list_psv = ["ezz", "ezr", "err", "ett"]
        name_list_sh = ["ezt", "ert"]
    elif output_type == "stress" or output_type == "stress_rate":
        name_list_psv = ["szz", "szr", "srr", "stt"]
        name_list_sh = ["szt", "srt"]
    elif output_type == "rota" or output_type == "rota_rate":
        name_list_psv = ["ot"]
        name_list_sh = ["oz", "or"]
    else:
        raise ValueError(
            "output_type must in  disp | velo | strain | strain_rate | "
            "stress | stress_rate | rota | rota_rate"
        )
    return name_list_psv, name_list_sh


def read_time_series_qseis2025_ascii(
        path_greenfunc, start_count, output_type="disp"
):
    time_seris_list = []
    start_count = start_count + 1
    name_list_psv, name_list_sh = get_outfile_name_list(output_type)
    for com in name_list_psv:
        ex_com = pd.read_csv(
            str(os.path.join(path_greenfunc, "ex.%s" % com)), sep="\\s+"
        ).to_numpy()  # type:ignore
        ss_com = pd.read_csv(
            str(os.path.join(path_greenfunc, "ss.%s" % com)), sep="\\s+"
        ).to_numpy()  # type:ignore
        ds_com = pd.read_csv(
            str(os.path.join(path_greenfunc, "ds.%s" % com)), sep="\\s+"
        ).to_numpy()  # type:ignore
        cl_com = pd.read_csv(
            str(os.path.join(path_greenfunc, "cl.%s" % com)), sep="\\s+"
        ).to_numpy()  # type:ignore
        time_series_com = np.concatenate(
            [
                ex_com[:, start_count],
                ss_com[:, start_count],
                ds_com[:, start_count],
                cl_com[:, start_count],
            ]
        ).reshape(4, -1)
        time_series_com = np.array(time_series_com, dtype=np.float32)
        time_seris_list.append(time_series_com)
    for com in name_list_sh:
        ss_r = pd.read_csv(
            os.path.join(path_greenfunc, "ss.%s" % com), sep="\\s+"
        ).to_numpy()  # type:ignore
        ds_r = pd.read_csv(
            os.path.join(path_greenfunc, "ds.%s" % com), sep="\\s+"
        ).to_numpy()  # type:ignore
        time_series_com = np.concatenate(
            [
                ss_r[:, start_count],
                ds_r[:, start_count],
            ]
        ).reshape(2, -1)
        time_series_com = np.array(time_series_com, dtype=np.float32)
        time_seris_list.append(time_series_com)
    return name_list_psv, name_list_sh, time_seris_list


def read_time_series_qseis2025_bin(
        path_greenfunc,
        start_count,
        output_type,
):
    time_series_list = []
    name_list_psv, name_list_sh = get_outfile_name_list(output_type)
    for com in name_list_psv:
        time_series_com = np.load(os.path.join(path_greenfunc, "grn_%s.npy" % com))
        time_series_list.append(time_series_com[:, start_count].reshape(4, -1))
    for com in name_list_sh:
        time_series_com = np.load(os.path.join(path_greenfunc, "grn_%s.npy" % com))
        time_series_list.append(time_series_com[:, start_count].reshape(2, -1))
    return name_list_psv, name_list_sh, time_series_list


def synthesize_rzv(time_series, m1):
    # ex,ss,ds,cl
    rzv = (
            time_series[0] * m1[0]
            + time_series[1] * m1[1]
            + time_series[2] * m1[2]
            + time_series[3] * m1[3]
    )
    return rzv


def synthesize_t(time_series, m2):
    t = time_series[0] * m2[0] + time_series[1] * m2[1]
    return t


def seek_qseis2025(
        path_green,
        event_depth_km,
        receiver_depth_km,
        az_deg,
        dist_km,
        focal_mechanism,
        srate,
        output_type,
        rotate=True,
        before_p=None,
        pad_zeros=False,
        shift=False,
        only_seismograms=True,
        model_name="ak135fc",
        green_info=None,
):
    if green_info is None:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            green_info = json.load(fr)
    srate_grn = 1 / green_info["sampling_interval"]
    sampling_num = green_info["sampling_num"]
    time_reduction_velo = green_info["time_reduction_velo"]
    dist_range = green_info["dist_range"]
    delta_dist = green_info["delta_dist"]
    num_each_group = green_info["N_each_group"]
    wavelet_type = green_info["wavelet_type"]
    grn_dep_list = green_info["event_depth_list"]
    grn_receiver_list = green_info["receiver_depth_list"]
    if not isinstance(grn_dep_list, list):
        grn_dep_source = grn_dep_list
    else:
        grn_dep_source = grn_dep_list[
            np.argmin(np.abs(event_depth_km - np.array(grn_dep_list)))
        ]
    if not isinstance(grn_receiver_list, list):
        grn_dep_receiver = grn_receiver_list
    else:
        grn_dep_receiver = grn_receiver_list[
            np.argmin(np.abs(receiver_depth_km - np.array(grn_receiver_list)))
        ]

    path_greenfunc = str(
        os.path.join(path_green, "%.2f" % grn_dep_source, "%.2f" % grn_dep_receiver)
    )
    ind = round((dist_km - dist_range[0]) / delta_dist)
    ind_group = ind // num_each_group
    grn_dist = dist_range[0] + ind * delta_dist
    start_count = ind - ind_group * num_each_group

    path_greenfunc_sub = os.path.join(path_greenfunc, "%d_0" % ind_group)
    if os.path.exists(os.path.join(path_greenfunc_sub, "grn_szt.npy")):
        name_list_psv, name_list_sh, time_series_list = (
            read_time_series_qseis2025_bin(
                path_greenfunc=path_greenfunc_sub,
                start_count=start_count,
                output_type=output_type,
            )
        )
    else:
        name_list_psv, name_list_sh, time_series_list = (
            read_time_series_qseis2025_ascii(
                path_greenfunc=path_greenfunc_sub,
                start_count=start_count,
                output_type=output_type,
            )
        )
    """
    Note 4:
    Double-Couple   m11/ m22/ m33/ m12/ m23/ m31  Azimuth_Factor_(tz,tr,tv)/(tt)
    ============================================================================
    explosion       1.0/ 1.0/ 1.0/ -- / -- / --       1.0         /   0.0
    strike-slip     -- / -- / -- / 1.0/ -- / --       sin(2*azi)  /   cos(2*azi)
                    1.0/-1.0/ -- / -- / -- / --       cos(2*azi)  /  -sin(2*azi)
    dip-slip        -- / -- / -- / -- / -- / 1.0      cos(azi)    /   sin(azi)
                    -- / -- / -- / -- / 1.0/ --       sin(azi)    /  -cos(azi)
    clvd           -0.5/-0.5/ 1.0/ -- / -- / --       1.0         /   0.0
    ============================================================================
    Single-Force    fx / fy / fz                  Azimuth_Factor_(tz,tr,tv)/(tt)
    ============================================================================
    fz              -- / -- / 1.0                        1.0      /   0.0
    fx              1.0/ -- / --                         cos(azi) /   sin(azi)
    fy              -- / 1.0/ --                         sin(azi) /  -cos(azi)
    ============================================================================
    """
    [M11, M12, M13, M22, M23, M33] = check_convert_fm(focal_mechanism=focal_mechanism)

    exp = (M11 + M22 + M33) / 3
    ss1 = M12
    ss2 = (M11 - M22) / 2
    ds1 = M13
    ds2 = M23
    clvd = (-M11 / 2 - M22 / 2 + M33) / (3 / 2)

    az_rad = np.deg2rad(az_deg)
    sin_az, cos_az = np.sin(az_rad), np.cos(az_rad)
    sin_2az, cos_2az = np.sin(2 * az_rad), np.cos(2 * az_rad)

    m1 = [exp, ss1 * sin_2az + ss2 * cos_2az, ds1 * cos_az + ds2 * sin_az, clvd]
    m2 = [ss1 * cos_2az - ss2 * sin_2az, ds1 * sin_az - ds2 * cos_az]

    # p-sv
    # ["tv,
    # "tr", "tz",
    # "ezz", "ezr", "err", "ett",
    # "szz", "szr", "srr", "stt",
    # "ot"]
    # sh
    # ["tt",
    # "ezt", "ert",
    # "szt", "srt",
    # "oz", "or"]
    if output_type in one_com_list:
        uv = synthesize_rzv(time_series=time_series_list[0], m1=m1)
        seismograms = np.array(uv)
    elif output_type in three_com_list:
        uz = -synthesize_rzv(time_series=time_series_list[0], m1=m1)
        ur = synthesize_rzv(time_series=time_series_list[1], m1=m1)
        ut = -synthesize_t(time_series=time_series_list[2], m2=m2)
        if rotate:
            seismograms = rotate_rtz_to_enz(az_deg, r=ur, t=ut, z=uz)
        else:
            seismograms = np.array([ur, ut, uz])
    elif output_type in rota_com_list:
        omega_t = -synthesize_rzv(time_series=time_series_list[0], m1=m1)
        omega_z = -synthesize_t(time_series=time_series_list[1], m2=m2)
        omega_r = synthesize_t(time_series=time_series_list[2], m2=m2)
        if rotate:
            seismograms = rotate_rtz_to_enz(az_deg, r=omega_r, t=omega_t, z=omega_z)
        else:
            seismograms = np.array([omega_r, omega_t, omega_z])
    elif output_type in six_com_list:
        s_zz = synthesize_rzv(time_series=time_series_list[0], m1=m1)
        s_zr = synthesize_rzv(time_series=time_series_list[1], m1=m1)
        s_rr = synthesize_rzv(time_series=time_series_list[2], m1=m1)
        s_tt = synthesize_rzv(time_series=time_series_list[3], m1=m1)
        s_zt = synthesize_t(time_series=time_series_list[4], m2=m2)
        s_rt = synthesize_t(time_series=time_series_list[5], m2=m2)

        sigma_tensor = np.array([s_tt, s_rt, -s_zt, s_rr, -s_zr, s_zz])
        if rotate:
            seismograms = rotate_symmetric_tensor_series(sigma_tensor.T, az_rad).T
        else:
            seismograms = sigma_tensor
    else:
        raise ValueError(
            "output_type must in  disp | velo | strain | strain_rate | "
            "stress | stress_rate | rota | rota_rate"
        )

    tpts_table = None
    if (before_p is not None) or shift or pad_zeros:
        first_p_grn, first_s_grn = read_tpts_table(
            path_green=path_green,
            event_depth_km=grn_dep_source,
            receiver_depth_km=grn_dep_receiver,
            ind=ind,
        )
        tpts_table = {"p_onset": first_p_grn, "s_onset": first_s_grn}

    if time_reduction_velo != 0:
        time_reduction = grn_dist / time_reduction_velo
    else:
        time_reduction = 0

    ts_count = 0
    if before_p is not None:
        ts_count = round(
            (tpts_table["p_onset"] - time_reduction - before_p) * srate_grn
        )
    if pad_zeros:
        if before_p is not None:
            raise ValueError("can not set before_p and pad_zeros together")
        ts_count = round(-time_reduction * srate_grn)
    seismograms = np.roll(seismograms, -ts_count)
    if ts_count > 0:
        seismograms[:, -ts_count:] = 0
    elif ts_count < 0:
        seismograms[:, :-ts_count] = 0

    first_p = None
    first_s = None
    if shift:
        seismograms, first_p, first_s = shift_green2real_tpts(
            seismograms=seismograms,
            tpts_table=tpts_table,
            srate=srate_grn,
            green_before_p=tpts_table["p_onset"] - time_reduction,
            event_depth_km=event_depth_km,
            dist_in_km=dist_km,
            receiver_depth_km=receiver_depth_km,
            model_name=model_name,
        )

    seismograms_resample = np.zeros(
        (len(seismograms), round(sampling_num * srate / srate_grn))
    )
    for i in range(len(seismograms)):
        seismograms_resample[i] = resample(
            seismograms[i], srate_old=srate_grn, srate_new=srate, zero_phase=True
        )
    if (wavelet_type == 1) and ("rate" not in output_type):
        seismograms_resample = np.cumsum(seismograms_resample, axis=1) / srate
    elif (wavelet_type == 2) and (("rate" in output_type) or (output_type == "velo")):
        seismograms_resample = (
                signal.convolve(
                    seismograms_resample.T,
                    np.array([1, -1])[:, None],
                    mode="same",
                    method="auto",
                ).T
                / srate
        )

    if only_seismograms:
        return seismograms_resample
    else:
        return (
            seismograms_resample,
            tpts_table,
            first_p,
            first_s,
            grn_dep_source,
            grn_dep_receiver,
            grn_dist,
        )


if __name__ == "__main__":
    pass



import matplotlib

matplotlib.use('tkagg')
import matplotlib.pyplot as plt
from scipy.signal import firwin, freqz, lfilter
import pandas as pd

from pygrnwang.read_qseis06_diff import *
from pygrnwang.read_qssp2020 import *
from pygrnwang.signal_process import resample, filter_butter

if __name__ == '__main__':
    event_depth_km = 10
    receiver_depth_km = 1
    az_deg = 0
    dist_km = 100
    srate = 4
    fm = [30, 40, 50]

    tp, ts = cal_first_p_s(
        event_depth_km, dist_km, receiver_depth_km)
    print(tp, ts)

    ind = 1
    cutoff = 0.25
    cut_length = 256

    numtaps = 21
    taps = firwin(numtaps, cutoff, fs=srate, window='hamming')

    strain_rate_qseis = seek_qseis06_strain_rate_diff(
        path_green="/e/grns_test/qseis06_strain",
        event_depth_km=event_depth_km,
        receiver_depth_km=receiver_depth_km,
        az_deg=az_deg,
        dist_km=dist_km,
        focal_mechanism=fm,
        srate=srate,
        before_p=None,
        pad_zeros=True,
        shift=False,
        only_seismograms=True,
        model_name="/e/grns_test/qseis06_strain/noq.nd",
        green_info=None,
    )

    strain_rate_qssp = seek_qssp2020(
        path_green='/e/grns_test/qssp_5_0.250_255.750_2.000_0.400_20000',
        event_depth_km=event_depth_km,
        receiver_depth_km=receiver_depth_km,
        az_deg=az_deg,
        dist_km=dist_km,
        focal_mechanism=fm,
        srate=srate,
        before_p=None,
        pad_zeros=True,
        shift=False,
        rotate=True,
        only_seismograms=True,
        output_type="strain_rate",
        model_name="/e/grns_test/qssp_5_0.250_255.750_2.000_0.400_20000/noq.nd",
        green_info=None,
    )

    # plt.figure()
    # plt.plot(taps)
    # plt.show()

    fig, axs = plt.subplots(nrows=6, ncols=1)
    for i in range(6):
        axs[i].plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
                    strain_rate_qseis[i][1:cut_length + 1])
        axs[i].plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
                    strain_rate_qssp[i][:cut_length])
        # axs[i].vlines(x=tp, ymin=np.min(strain_rate_qseis[:cut_length]),
        #               ymax=np.max(strain_rate_qseis[:cut_length]), color='red')
    axs[-1].legend(['qseis', 'qssp'])
    plt.xlabel('Time (s)')
    plt.show()

    fig, axs = plt.subplots(nrows=6, ncols=1)
    for i in range(6):
        strain_rate_qseis_conv = np.convolve(strain_rate_qseis[i], taps, mode='full')
        strain_rate_qssp_conv = np.convolve(strain_rate_qssp[i], taps, mode='full')
        axs[i].plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
                    strain_rate_qseis_conv[1:cut_length + 1])
        axs[i].plot(np.linspace(0, cut_length / srate, cut_length, endpoint=False),
                    strain_rate_qssp_conv[:cut_length])
        # axs[i].vlines(x=tp, ymin=np.min(strain_rate_qseis[:cut_length]),
        #               ymax=np.max(strain_rate_qseis[:cut_length]), color='red')
    axs[-1].legend(['qseis', 'qssp'])
    plt.xlabel('Time (s)')
    plt.show()



import math
import warnings

import numpy as np

# a: Radius of Earth in m. Uses the value for WGS84 by default.
# f: Flattening of Earth. Uses the value for WGS84 by default.
WGS84_A = 6378137.0
WGS84_F = 1 / 298.257223563


def to_int_or_zero(value):
    """
    Converts given value to an integer or returns 0 if it fails.

    :param value: Arbitrary data type.
    :rtype: int

    .. rubric:: Example

    >>> to_int_or_zero("12")
    12

    >>> to_int_or_zero("x")
    0
    """
    try:
        return int(value)
    except ValueError:
        return 0


# checking for geographiclib
try:
    import geographiclib  # @UnusedImport # NOQA
    from geographiclib.geodesic import Geodesic

    HAS_GEOGRAPHICLIB = True
    try:
        GEOGRAPHICLIB_VERSION_AT_LEAST_1_34 = [1, 34] <= list(
            map(to_int_or_zero, geographiclib.__version__.split("."))
        )
    except AttributeError:
        GEOGRAPHICLIB_VERSION_AT_LEAST_1_34 = False
except ImportError:
    HAS_GEOGRAPHICLIB = False
    GEOGRAPHICLIB_VERSION_AT_LEAST_1_34 = False


def _check_latitude(latitude, variable_name="latitude"):
    """
    Check whether latitude is in the -90 to +90 range.
    """
    if latitude is None:
        return
    if latitude > 90 or latitude < -90:
        msg = "{} out of bounds! (-90 <= {} <=90)".format(variable_name, variable_name)
        raise ValueError(msg)


def _normalize_longitude(longitude):
    """
    Normalize longitude in the -180 to +180 range.
    """
    if longitude is None:
        return
    while longitude > 180:
        longitude -= 360
    while longitude < -180:
        longitude += 360
    return longitude


def calc_vincenty_inverse_obspy(lat1, lon1, lat2, lon2, a=WGS84_A, f=WGS84_F):
    """
    Vincenty Inverse Solution of Geodesics on the Ellipsoid.

    Computes the distance between two geographic points on the WGS84
    ellipsoid and the forward and backward azimuths between these points.

    :param lat1: Latitude of point A in degrees (positive for northern,
        negative for southern hemisphere)
    :param lon1: Longitude of point A in degrees (positive for eastern,
        negative for western hemisphere)
    :param lat2: Latitude of point B in degrees (positive for northern,
        negative for southern hemisphere)
    :param lon2: Longitude of point B in degrees (positive for eastern,
        negative for western hemisphere)
    :param a: Radius of Earth in m. Uses the value for WGS84 by default.
    :param f: Flattening of Earth. Uses the value for WGS84 by default.
    :return: (Great circle distance in m, azimuth A->B in degrees,
        azimuth B->A in degrees)
    :raises: This method may have no solution between two nearly antipodal
        points; an iteration limit traps this case and a ``StopIteration``
        exception will be raised.

    .. note::
        This code is based on an implementation incorporated in
        Matplotlib Basemap Toolkit 0.9.5 http://sourceforge.net/projects/\
matplotlib/files/matplotlib-toolkits/basemap-0.9.5/
        (matplotlib/toolkits/basemap/greatcircle.py)

        Algorithm from Geocentric Datum of Australia Technical Manual.

        * https://www.icsm.gov.au/publications
        * https://www.icsm.gov.au/publications/gda2020-technical-manual-v16

        It states::

            There are a number of formulae that are available to calculate
            accurate geodetic positions, azimuths and distances on the
            ellipsoid.

            Vincenty's formulae (Vincenty, 1975) may be used for lines ranging
            from a few cm to nearly 20,000 km, with millimetre accuracy. The
            formulae have been extensively tested for the Australian region, by
            comparison with results from other formulae (Rainsford, 1955 &
            Sodano, 1965).

            * Inverse problem: azimuth and distance from known latitudes and
              longitudes
            * Direct problem: Latitude and longitude from known position,
              azimuth and distance.
    """
    # Check inputs
    _check_latitude(lat1, "lat1")
    lon1 = _normalize_longitude(lon1)
    _check_latitude(lat2, "lat2")
    lon2 = _normalize_longitude(lon2)

    b = a * (1 - f)  # semiminor axis

    if math.isclose(lat1, lat2) and math.isclose(lon1, lon2):
        return 0.0, 0.0, 0.0

    # convert latitudes and longitudes to radians:
    lat1 = math.radians(lat1)
    lon1 = math.radians(lon1)
    lat2 = math.radians(lat2)
    lon2 = math.radians(lon2)

    tan_u1 = (1 - f) * math.tan(lat1)
    tan_u2 = (1 - f) * math.tan(lat2)

    u_1 = math.atan(tan_u1)
    u_2 = math.atan(tan_u2)

    dlon = lon2 - lon1
    last_dlon = -4000000.0  # an impossible value
    omega = dlon

    # Iterate until no significant change in dlon or iterlimit has been
    # reached (http://www.movable-type.co.uk/scripts/latlong-vincenty.html)
    iterlimit = 100
    try:
        while (
            last_dlon < -3000000.0
            or dlon != 0
            and abs((last_dlon - dlon) / dlon) > 1.0e-9
        ):
            sqr_sin_sigma = pow(math.cos(u_2) * math.sin(dlon), 2) + pow(
                (
                    math.cos(u_1) * math.sin(u_2)
                    - math.sin(u_1) * math.cos(u_2) * math.cos(dlon)
                ),
                2,
            )
            sin_sigma = math.sqrt(sqr_sin_sigma)

            cos_sigma = math.sin(u_1) * math.sin(u_2) + math.cos(u_1) * math.cos(
                u_2
            ) * math.cos(dlon)
            sigma = math.atan2(sin_sigma, cos_sigma)
            sin_alpha = math.cos(u_1) * math.cos(u_2) * math.sin(dlon) / sin_sigma

            sqr_cos_alpha = 1 - sin_alpha * sin_alpha
            if math.isclose(sqr_cos_alpha, 0):
                # Equatorial line
                cos2sigma_m = 0
            else:
                cos2sigma_m = cos_sigma - (
                    2 * math.sin(u_1) * math.sin(u_2) / sqr_cos_alpha
                )

            c = (f / 16) * sqr_cos_alpha * (4 + f * (4 - 3 * sqr_cos_alpha))
            last_dlon = dlon
            dlon = omega + (1 - c) * f * sin_alpha * (
                sigma
                + c
                * sin_sigma
                * (cos2sigma_m + c * cos_sigma * (-1 + 2 * pow(cos2sigma_m, 2)))
            )

            iterlimit -= 1
            if iterlimit < 0:
                # iteration limit reached
                raise StopIteration
    except ValueError:
        # usually "math domain error"
        raise StopIteration

    u2 = sqr_cos_alpha * (a * a - b * b) / (b * b)
    _a = 1 + (u2 / 16384) * (4096 + u2 * (-768 + u2 * (320 - 175 * u2)))
    _b = (u2 / 1024) * (256 + u2 * (-128 + u2 * (74 - 47 * u2)))
    delta_sigma = (
        _b
        * sin_sigma
        * (
            cos2sigma_m
            + (_b / 4)
            * (
                cos_sigma * (-1 + 2 * pow(cos2sigma_m, 2))
                - (_b / 6)
                * cos2sigma_m
                * (-3 + 4 * sqr_sin_sigma)
                * (-3 + 4 * pow(cos2sigma_m, 2))
            )
        )
    )

    dist = b * _a * (sigma - delta_sigma)
    alpha12 = math.atan2(
        (math.cos(u_2) * math.sin(dlon)),
        (
            math.cos(u_1) * math.sin(u_2)
            - math.sin(u_1) * math.cos(u_2) * math.cos(dlon)
        ),
    )
    alpha21 = math.atan2(
        (math.cos(u_1) * math.sin(dlon)),
        (
            -math.sin(u_1) * math.cos(u_2)
            + math.cos(u_1) * math.sin(u_2) * math.cos(dlon)
        ),
    )

    if alpha12 < 0.0:
        alpha12 = alpha12 + (2.0 * math.pi)
    if alpha12 > (2.0 * math.pi):
        alpha12 = alpha12 - (2.0 * math.pi)

    alpha21 = alpha21 + math.pi

    if alpha21 < 0.0:
        alpha21 = alpha21 + (2.0 * math.pi)
    if alpha21 > (2.0 * math.pi):
        alpha21 = alpha21 - (2.0 * math.pi)

    # convert to degrees:
    alpha12 = alpha12 * 360 / (2.0 * math.pi)
    alpha21 = alpha21 * 360 / (2.0 * math.pi)

    return dist, alpha12, alpha21


def gps2dist_azimuth(lat1, lon1, lat2, lon2, a=WGS84_A, f=WGS84_F):
    """
    Computes the distance between two geographic points on the WGS84
    ellipsoid and the forward and backward azimuths between these points.

    :param lat1: Latitude of point A in degrees (positive for northern,
        negative for southern hemisphere)
    :param lon1: Longitude of point A in degrees (positive for eastern,
        negative for western hemisphere)
    :param lat2: Latitude of point B in degrees (positive for northern,
        negative for southern hemisphere)
    :param lon2: Longitude of point B in degrees (positive for eastern,
        negative for western hemisphere)
    :param a: Radius of Earth in m. Uses the value for WGS84 by default.
    :param f: Flattening of Earth. Uses the value for WGS84 by default.
    :return: (Great circle distance in m, azimuth A->B in degrees,
        azimuth B->A in degrees)

    .. note::
        This function will check if you have installed the Python module
        `geographiclib <http://geographiclib.sf.net>`_ - a very fast module
        for converting between geographic, UTM, UPS, MGRS, and geocentric
        coordinates, for geoid calculations, and for solving geodesic problems.
        Otherwise the locally implemented Vincenty's Inverse formulae
        (:func:`obspy.geodetics.base.calc_vincenty_inverse`) is used which
        has known limitations for two nearly antipodal points and is ca. 4x
        slower.
    """
    if HAS_GEOGRAPHICLIB:
        _check_latitude(lat1, "lat1")
        _check_latitude(lat2, "lat2")
        result = Geodesic(a=a, f=f).Inverse(lat1, lon1, lat2, lon2)
        azim = result["azi1"]
        if azim < 0:
            azim += 360
        bazim = result["azi2"] + 180
        return (result["s12"], azim, bazim)
    else:
        try:
            values = calc_vincenty_inverse_obspy(lat1, lon1, lat2, lon2, a, f)
            if np.all(np.isnan(values)):
                raise StopIteration
            return values
        except StopIteration:
            msg = (
                "Catching unstable calculation on antipodes. "
                "The currently used Vincenty's Inverse formulae "
                "has known limitations for two nearly antipodal points. "
                "Install the Python module 'geographiclib' to solve this "
                "issue."
            )
            warnings.warn(msg)
            return (20004314.5, 0.0, 0.0)
        except ValueError as e:
            raise e



import os
import math

from .edcmp2inp import s as str_inp
from .utils import call_exe


def create_inp_edcmp2(
    path_green,
    obs_depth,
    obs_x_range,
    obs_y_range,
    obs_delta_x,
    obs_delta_y,
    source_array_edcmp,
    layered=True,
    lam=30516224000,
    mu=33701888000,
):
    """

    :param path_green:
    :param obs_depth: km
    :param obs_x_range: km
    :param obs_y_range: km
    :param obs_delta_x: km
    :param obs_delta_y: km
    :param source_array_edcmp:
     [slip(m), x(km), y(km), z(km),
     strike(deg), dip(deg), rake(deg),
     sub_len_strike(km), sub_len_dip(km)]
    :param layered:
    :param lam:
    :param mu:
    :return:
    """
    lines = str_inp.split("\n")
    lines = [line + "\n" for line in lines]
    lines_after_sources = lines[97:]
    lines = lines[:96]

    nx = math.ceil((obs_x_range[1] - obs_x_range[0]) / obs_delta_x) + 1
    lines[45] = "%d %f %f\n" % (
        nx,
        obs_x_range[0] * 1e3,
        obs_x_range[1] * 1e3,
    )
    ny = math.ceil((obs_y_range[1] - obs_y_range[0]) / obs_delta_y) + 1
    lines[46] = "%d %f %f\n" % (
        ny,
        obs_y_range[0] * 1e3,
        obs_y_range[1] * 1e3,
    )
    path_edcmp_obs_dep = str(os.path.join(path_green, "edcmp2", "%.2f" % obs_depth, ""))
    lines[59] = "'%s'\n" % path_edcmp_obs_dep
    n_sources = len(source_array_edcmp)
    lines[91] = "%d\n" % n_sources

    lines_sources = []
    for i in range(len(source_array_edcmp)):
        si = "%d %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f\n" % (
            i + 1,
            source_array_edcmp[i, 0],
            source_array_edcmp[i, 1] * 1e3,
            source_array_edcmp[i, 2] * 1e3,
            source_array_edcmp[i, 3] * 1e3,
            source_array_edcmp[i, 7] * 1e3,
            source_array_edcmp[i, 8] * 1e3,
            source_array_edcmp[i, 4],
            source_array_edcmp[i, 5],
            source_array_edcmp[i, 6],
        )
        lines_sources.append(si)

    if layered:
        lines_after_sources[32] = "1\n"
        path_edgrn = os.path.join(path_green, "edgrn2", "%.2f" % obs_depth, "")
        lines_after_sources[33] = "'%s' 'edgrn.ss' 'edgrn.ds' 'edgrn.cl'\n" % (
            path_edgrn
        )
    else:
        lines_after_sources[32] = "0\n"
        lines_after_sources[33] = "%f %f %f\n" % (obs_depth * 1e3, lam, mu)

    lines = lines + lines_sources + lines_after_sources
    with open(os.path.join(path_edcmp_obs_dep, "grn.inp"), "w") as fw:
        fw.writelines(lines)


def call_edcmp2(obs_depth, path_green, check_finished=False):
    os.chdir(path_green)
    sub_sub_dir = str(os.path.join("edcmp2", "%.2f" % obs_depth))
    if (
        check_finished
        and os.path.exists(os.path.join(sub_sub_dir, ".finished"))
        and len(os.listdir(sub_sub_dir)) > 2
    ):
        return None
    path_inp = str(os.path.join(sub_sub_dir, "grn.inp"))
    path_finished = os.path.join(sub_sub_dir, ".finished")

    if (
        check_finished
        and os.path.exists(path_finished)
        and len(os.listdir(sub_sub_dir)) > 2
    ):
        with open(path_finished, "r") as fr:
            output = fr.readlines()
        return output

    output = call_exe(
        path_green=path_green,
        path_inp=path_inp,
        path_finished=path_finished,
        name="edcmp2",
    )
    return output



import os

import numpy as np


def read_green_info_spgrn2020(path_greenfunc: str, green_depth: float) -> dict:
    """
    read GreenInfo.dat
    :param path_greenfunc:
    :param green_depth: unit km
    :return: {
             "time_window": time_window, type float, unit s
             "sampling_interval": sampling_interval, type float, unit s
             "samples_num": int(samples_num), type int
             "dist_list": dist_list, type list, unit km
         }
    """
    with open(
        os.path.join(path_greenfunc, "GreenInfo%.2f.dat" % green_depth), "r"
    ) as fr:
        lines = fr.readlines()
    [time_window, sampling_interval, samples_num] = lines[6].strip().split()
    time_window = float(time_window)
    sampling_interval = float(sampling_interval)
    samples_num = int(samples_num)
    number_of_distance = float(lines[9].strip())
    dist_list = []
    for i in range(round(np.ceil(number_of_distance / 5))):
        temp = lines[10 + i].strip().split()
        for item in temp:
            dist_list.append(float(item))
    return {
        "time_window": time_window,
        "sampling_interval": sampling_interval,
        "samples_num": int(samples_num),
        "dist_list": dist_list,
    }



import os
import platform
import subprocess

import numpy as np
import pandas as pd

from .signal_process import linear_interp
from .pytaup import cal_first_p_s


def read_source_array(source_inds, path_input, shift2corner=False, source_shapes=None):
    source_array = None
    for ind_src in range(len(source_inds)):
        source_plane = pd.read_csv(
            str(os.path.join(path_input, "source_plane%d.csv" % source_inds[ind_src])),
            index_col=False,
            header=None,
        ).to_numpy()
        if shift2corner:
            mu_strike = (
                source_plane[source_shapes[ind_src][1], :3] - source_plane[0, :3]
            )
            mu_dip = source_plane[1, :3] - source_plane[0, :3]
            source_plane[:, :3] = source_plane[:, :3] - mu_strike / 2 - mu_dip / 2
        if ind_src == 0:
            source_array = source_plane.copy()
        else:
            source_array = np.concatenate([source_array, source_plane.copy()], axis=0)
    return source_array


def group(inp_list, num_in_each_group):
    group_list = []
    for i in range(len(inp_list) // num_in_each_group):
        group_list.append(inp_list[i * num_in_each_group : (i + 1) * num_in_each_group])
    rest = len(inp_list) % num_in_each_group
    if rest != 0:
        group_list.append(inp_list[-rest:])
    return group_list


def shift_green2real_tpts(
    seismograms,
    tpts_table,
    green_before_p,
    srate,
    event_depth_km,
    dist_in_km,
    receiver_depth_km=0,
    model_name="ak135",
):
    first_p, first_s = cal_first_p_s(
        event_depth_km=event_depth_km,
        dist_km=dist_in_km,
        receiver_depth_km=receiver_depth_km,
        model_name=model_name,
    )
    p_count = round(green_before_p * srate)
    s_count = round(
        (tpts_table["s_onset"] - tpts_table["p_onset"] + green_before_p) * srate
    )
    p_count_new = round((first_p - tpts_table["p_onset"] + green_before_p) * srate)
    s_count_new = min(
        len(seismograms[0]),
        round((first_s - tpts_table["p_onset"] + green_before_p) * srate),
    )
    for i in range(seismograms.shape[0]):
        green_before_p = seismograms[i][:p_count]
        p_s = linear_interp(seismograms[i][p_count:s_count], s_count_new - p_count_new)
        after_s = seismograms[i][s_count:]
        if len(after_s) > 0:
            after_s = linear_interp(
                after_s, len(seismograms[i]) - len(green_before_p) - len(p_s)
            )
            seismograms[i] = np.concatenate([green_before_p, p_s, after_s])
        else:
            seismograms[i] = np.concatenate([green_before_p, p_s])[
                : len(seismograms[i])
            ]

    return seismograms, first_p, first_s


def cal_max_dist_from_2d_points(A: np.ndarray, B: np.ndarray):
    """

    :param A: (m,2)
    :param B: (n,2)
    :return: max_distance
    """
    # Calculate the differences in each dimension (broadcasting)
    differences = A[:, np.newaxis, :] - B[np.newaxis, :, :]

    # Square the differences and sum across columns (to get squared distances)
    squared_distances = np.sum(differences**2, axis=2)

    # Take the square root to get Euclidean distances
    distances = np.sqrt(squared_distances)

    # Find the maximum distance
    max_distance = np.max(distances)
    return max_distance


def create_rotate_z_mat(gamma):
    """
    Generates a rotation matrix about the Z-axis.

    Parameters:
        gamma : float
            Rotation angle in radians.

    Returns:
        R : numpy.ndarray
            A 3x3 rotation matrix.
    """
    R = np.array(
        [
            [np.cos(gamma), -np.sin(gamma), 0],
            [np.sin(gamma), np.cos(gamma), 0],
            [0, 0, 1],
        ]
    )
    return R


def rotate_symmetric_tensor_series(tensor, gamma):
    """
    Rotates a series of symmetric tensors without using an explicit loop.

    Parameters:
        tensor: numpy array of shape (n, 6)
            Each row is [xx, xy, xz, yy, yz, zz] representing a symmetric tensor.
        gamma: float
            Rotation angle (in radians) used to create the rotation matrix.

    Returns:
        rotated_tensor: numpy array of shape (n, 6)
            Rotated tensor components in the same order as the input.
    """
    # Create the 3x3 rotation matrix (assumed to be defined elsewhere).
    R = create_rotate_z_mat(gamma)
    n = tensor.shape[0]

    # Construct full symmetric matrices from the condensed tensor representation.
    A = np.empty((n, 3, 3), dtype=tensor.dtype)
    A[:, 0, 0] = tensor[:, 0]
    A[:, 0, 1] = tensor[:, 1]
    A[:, 0, 2] = tensor[:, 2]
    A[:, 1, 0] = tensor[:, 1]
    A[:, 1, 1] = tensor[:, 3]
    A[:, 1, 2] = tensor[:, 4]
    A[:, 2, 0] = tensor[:, 2]
    A[:, 2, 1] = tensor[:, 4]
    A[:, 2, 2] = tensor[:, 5]

    # Rotate each tensor using batch matrix multiplication:
    # Compute rotated_A = R.T @ A @ R for each tensor.
    rotated_A = np.einsum("ij,njk,kl->nil", R.T, A, R)

    # Extract the independent components from the rotated tensors.
    rotated_tensor = np.empty((n, 6), dtype=tensor.dtype)
    rotated_tensor[:, 0] = rotated_A[:, 0, 0]
    rotated_tensor[:, 1] = rotated_A[:, 0, 1]
    rotated_tensor[:, 2] = rotated_A[:, 0, 2]
    rotated_tensor[:, 3] = rotated_A[:, 1, 1]
    rotated_tensor[:, 4] = rotated_A[:, 1, 2]
    rotated_tensor[:, 5] = rotated_A[:, 2, 2]

    return rotated_tensor


def convert_earth_model_nd2inp(path_nd, path_output):
    with open(path_nd, "r") as fr:
        lines = fr.readlines()
    lines_new = []
    for i in range(len(lines)):
        temp = lines[i].split()
        if len(temp) > 1:
            lines_new.append(temp)
    for i in range(len(lines_new)):
        # print(lines_new[i])
        lines_new[i] = "  ".join([str(int(i + 1))] + lines_new[i]) + "\n"  # type:ignore
    # with open(path_output, "w") as fw:
    #     fw.writelines(lines_new)
    return lines_new


def convert_earth_model_nd2nd_without_Q(path_nd, path_output):
    with open(path_nd, "r") as fr:
        lines = fr.readlines()
    lines_new = []
    for i in range(len(lines)):
        temp = lines[i].split()
        if len(temp) > 1:
            lines_new.append([str(float(_)) for _ in temp[:-2]])
            lines_new[i] = "  ".join(lines_new[i]) + "\n"
        else:
            lines_new.append(lines[i].strip() + "\n")
    with open(path_output, "w") as fw:
        fw.writelines(lines_new)
    return lines_new


def read_nd(path_nd):
    with open(path_nd, "r") as fr:
        lines = fr.readlines()
    lines_new = []
    for i in range(len(lines)):
        temp = lines[i].split()
        if len(temp) > 1:
            for j in range(len(temp)):
                lines_new.append(float(temp[j]))
    nd_model = np.array(lines_new).reshape(-1, 4)
    return nd_model


def read_material_nd(model_name, depth):
    if model_name == "ak135fc":
        from .ak135fc import s as str_nd

        lines = str_nd.split("\n")
        lines_new = []
        for i in range(len(lines)):
            temp = lines[i].split()
            if len(temp) > 1:
                for j in range(len(temp)):
                    lines_new.append(float(temp[j]))
        nd_model = np.array(lines_new).reshape(-1, 4)
    else:
        nd_model = read_nd(model_name)
    ind = np.argwhere((nd_model[:, 0] - depth) >= 0)[0][0]
    return nd_model[ind]


def read_layerd_material(path_layerd_dat, depth_in_km):
    # thickness, rho, vp, vs, qp, qs
    depth_in_m = depth_in_km * 1e3
    dat = np.loadtxt(path_layerd_dat)
    ind = np.argwhere((np.cumsum(dat[:, 0]) - depth_in_m) >= 0)[0][0]
    return dat[ind]


def create_stf(tau, srate):
    t = np.linspace(0, tau, round(tau * srate) + 1, endpoint=True)
    stf = (2 / tau) * (np.sin(np.pi * t / tau)) ** 2
    return stf


def group_planes(strike_array):
    """
    It is necessary to ensure that the sub faults on
    the same fault plane have the same strikes!!!
    :param strike_array: numpy array

    Returns:
    np.array: An array containing the lengths of each group.
    """
    # Find the indices where the value changes
    # a[1:] != a[:-1] produces a boolean array that's True
    # at positions where a value differs from its predecessor.
    change_indices = np.where(strike_array[1:] != strike_array[:-1])[0] + 1

    # Include the start and end indices to get boundaries for each group.
    boundaries = np.concatenate(([0], change_indices, [len(strike_array)]))

    # The difference between consecutive boundaries gives the group lengths.
    lengths = np.diff(boundaries)
    return lengths


def reshape_sub_faults(sub_faults, num_strike, num_dip):
    mu_strike = sub_faults[num_dip] - sub_faults[0]
    mu_dip = sub_faults[1] - sub_faults[0]
    sub_faults = sub_faults - mu_strike / 2 - mu_dip / 2
    X: np.ndarray = sub_faults[:, 0]
    Y: np.ndarray = sub_faults[:, 1]
    Z: np.ndarray = sub_faults[:, 2]

    X = X.reshape(num_strike, num_dip)
    Y = Y.reshape(num_strike, num_dip)
    Z = Z.reshape(num_strike, num_dip)

    X = np.concatenate([X, np.array([X[:, -1] + mu_dip[0]]).T], axis=1)
    Y = np.concatenate([Y, np.array([Y[:, -1] + mu_dip[1]]).T], axis=1)
    Z = np.concatenate([Z, np.array([Z[:, -1] + mu_dip[2]]).T], axis=1)

    X = np.concatenate([X, np.array([X[-1, :] + mu_strike[0]])], axis=0)
    Y = np.concatenate([Y, np.array([Y[-1, :] + mu_strike[1]])], axis=0)
    Z = np.concatenate([Z, np.array([Z[-1, :] + mu_strike[2]])], axis=0)
    return X, Y, Z


def call_exe(path_green, path_inp, path_finished, name):
    name_exe = "%s.exe" % name if platform.system() == "Windows" else "%s.bin" % name
    path_exe = os.path.join(path_green, name_exe)
    proc = subprocess.Popen(
        [path_exe],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout_bytes, stderr_bytes = proc.communicate(str.encode(path_inp))
    stdout_text = stdout_bytes.decode(errors="ignore")
    stderr_text = stderr_bytes.decode(errors="ignore")
    output = stdout_text + stderr_text
    with open(path_finished, "w") as fw:
        fw.writelines(output)
        return None


if __name__ == "__main__":
    pass



import os
import shutil
import platform
import pickle
import json
import datetime
import math
from multiprocessing import Pool

import numpy as np
from tqdm import tqdm

try:
    from mpi4py import MPI
except:
    pass

from .create_edcmp import create_inp_edcmp2, call_edcmp2
from .utils import group
from .geo import d2km, convert_sub_faults_geo2ned, cal_max_dist_from_2d_points


def _call_edcmp2_star(args):
    return call_edcmp2(*args)


def pre_process_edcmp2(
    processes_num,
    path_green,
    path_bin,
    obs_depth_list,
    obs_x_range,
    obs_y_range,
    obs_delta_x,
    obs_delta_y,
    source_array,
    source_ref=None,
    obs_ref=None,
    layered=True,
    lam=30516224000,
    mu=33701888000,
):
    """
    Pre-processes input data for edcmp2, updates Green's function library settings,
    and prepares input files for each observation depth.

    :param processes_num: Number of parallel processes to be used.
    :param path_green: Directory path for the Green's function library.
    :param path_bin: Path to the edcmp2 binary file to be used if not present in path_green.
    :param obs_depth_list: List of observation depths (in km) for which inputs are to be created.
    :param obs_x_range: List or tuple [min, max] of observation x-coordinates (latitude in degrees).
    :param obs_y_range: List or tuple [min, max] of observation y-coordinates (longitude in degrees).
    :param obs_delta_x: Sampling interval in the x-direction (in degrees).
    :param obs_delta_y: Sampling interval in the y-direction (in degrees).
    :param source_array: NumPy array with each row containing source parameters:
                         [slip (m), lat (deg), lon (deg), depth (km), strike (deg), dip (deg), rake (deg)].
    :param source_ref: Optional reference coordinate for sources [lat, lon]. If provided,
                       source coordinates will be converted from geographic to NED coordinates.
    :param obs_ref: Optional reference coordinate for observations [lat, lon]. If provided,
                    the observation coordinate ranges and intervals are converted to km.
    :param layered: Boolean flag indicating whether the Green's function library is for a layered model.
    :param lam: Lamé parameter λ (default: 30516224000).
    :param mu: Shear modulus μ (default: 33701888000).
    :return: group_list_edcmp, a grouping of observation depths according to processes_num,
             which is also saved to a pickle file in path_green.
    """
    # Print status (commented out)
    # print("preprocessing edcmp2")

    # Create a copy of the source array.
    source_array_new = source_array.copy()

    # Determine the correct binary filename based on the operating system.
    if platform.system() == "Windows":
        path_bin_call = os.path.join(path_green, "edcmp2.exe")
    else:
        path_bin_call = os.path.join(path_green, "edcmp2.bin")
    shutil.copy(path_bin, path_bin_call)

    # Load the current Green's function library information.
    with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
        green_info = json.load(fr)

    # If a source reference is provided, convert source coordinates.
    if source_ref is not None:
        sources_ned = convert_sub_faults_geo2ned(
            source_array_new[:, 1:4], np.concatenate([source_ref, np.zeros(1)]), True
        )
        source_array_new[:, 1:4] = (
            sources_ned / 1e3
        )  # Convert NED coordinates back to km.

    # If an observation reference is provided, convert observation ranges and intervals.
    if obs_ref is not None:
        obs_x_range = list(
            (np.array(obs_x_range) - obs_ref[0]) * d2km
        )  # Convert x-range to km.
        obs_y_range = list(
            (np.array(obs_y_range) - obs_ref[1]) * d2km
        )  # Convert y-range to km.
        obs_delta_x = obs_delta_x * d2km  # Convert x sampling interval to km.
        obs_delta_y = obs_delta_y * d2km  # Convert y sampling interval to km.

    # Check if the observation grid exceeds the range supported by the Green's function library.
    nx = math.ceil((obs_x_range[1] - obs_x_range[0]) / obs_delta_x) + 1
    ny = math.ceil((obs_y_range[1] - obs_y_range[0]) / obs_delta_y) + 1
    obs_points_x, obs_points_y = np.meshgrid(
        np.linspace(obs_x_range[0], obs_x_range[1], nx),
        np.linspace(obs_y_range[0], obs_y_range[1], ny),
    )
    obs_points = np.concatenate(
        [np.array([obs_points_x.flatten()]).T, np.array([obs_points_y.flatten()]).T],
        axis=1,
    )
    max_distance = cal_max_dist_from_2d_points(source_array_new[:, 1:3], obs_points)
    if max_distance > green_info["grn_dist_range"][1]:
        raise ValueError(
            "\nThe max distance in obs is %f km,\n"
            "exceeds the max value %f km\n"
            "in the Green's function library!"
            % (max_distance, green_info["grn_dist_range"][1])
        )

    # For each observation depth, create the corresponding edcmp2 input file.
    for obs_depth in obs_depth_list:
        sub_sub_dir = str(os.path.join(path_green, "edcmp2", "%.2f" % obs_depth))
        os.makedirs(sub_sub_dir, exist_ok=True)
        create_inp_edcmp2(
            path_green,
            obs_depth,
            obs_x_range,
            obs_y_range,
            obs_delta_x,
            obs_delta_y,
            source_array_new,
            layered,
            lam,
            mu,
        )

    # Update the green_info dictionary with new observation and model parameters.
    green_info["obs_x_range"] = obs_x_range
    green_info["obs_y_range"] = obs_y_range
    green_info["obs_delta_x"] = obs_delta_x
    green_info["obs_delta_y"] = obs_delta_y
    green_info["source_ref"] = source_ref
    green_info["obs_ref"] = obs_ref
    green_info["layered"] = layered
    green_info["lam"] = lam
    green_info["mu"] = mu
    json_str = json.dumps(green_info, indent=4, ensure_ascii=False)
    with open(
        os.path.join(path_green, "green_lib_info.json"), "w", encoding="utf-8"
    ) as file:
        file.write(json_str)

    # Group the observation depths into subgroups for parallel processing.
    group_list_edcmp = group(obs_depth_list, processes_num)
    with open(os.path.join(path_green, "group_list_edcmp.pkl"), "wb") as fw:
        pickle.dump(group_list_edcmp, fw)  # type: ignore

    return group_list_edcmp


def compute_static_stress_edcmp2_sequential(path_green, check_finished=False):
    # s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list_edcmp.pkl"), "rb") as fr:
        group_list_edcmp = pickle.load(fr)
    for item in tqdm(group_list_edcmp, desc="Computing static stress"):
        for i in range(len(item)):
            # print("computing " + str(item[i]) + " km")
            call_edcmp2(item[i], path_green, check_finished)
    # e = datetime.datetime.now()
    # print("run time:%s" % str(e - s))


def compute_static_stress_edcmp2_parallel(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list_edcmp.pkl"), "rb") as fr:
        group_list_edcmp = pickle.load(fr)
    tasks = []
    for grp in group_list_edcmp:
        for d in grp:
            tasks.append((d, path_green, check_finished))

    processes = None
    try:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            processes = json.load(fr).get("processes_num", None)
    except Exception:
        pass

    with Pool(processes=processes) as pool:
        for _ in tqdm(
            pool.imap_unordered(_call_edcmp2_star, tasks, chunksize=1),
            total=len(tasks),
            desc="Computing static stress",
        ):
            pass
    e = datetime.datetime.now()
    return e - s


def create_grnlib_edcmp2_parallel_multi_nodes(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list_edcmp.pkl"), "rb") as fr:
        group_list_edcmp = pickle.load(fr)
    for ind_group in range(len(group_list_edcmp)):
        comm = MPI.COMM_WORLD
        processes_num = comm.Get_size()
        rank = comm.Get_rank()
        if processes_num != len(group_list_edcmp[0]):
            raise ValueError(
                "processes_num is %d, item num in group is %d. \n"
                "Pleasse check the process num!"
                % (processes_num, len(group_list_edcmp[0]))
            )
        print("ind_group:%d rank:%d" % (ind_group, rank))
        call_edcmp2(
            obs_depth=group_list_edcmp[ind_group][rank],
            path_green=path_green,
            check_finished=check_finished,
        )
    e = datetime.datetime.now()
    print("run time:" + str(e - s))



import os
import platform
import shutil
import subprocess
from setuptools import setup
from setuptools.command.build_py import build_py as _build_py
from setuptools import setup, Distribution

project_root = os.path.dirname(os.path.abspath(__file__))
platform_exec = "exe" if platform.system() == "Windows" else "bin"


class BinaryDistribution(Distribution):
    def has_ext_modules(self):
        return True


def _compile_dir(src_dir: str, out_bin: str, extra_flags: list[str]) -> None:
    cmd = "gfortran ./*.f -O3 %s -o %s" % (
        " ".join("%s" % extra_flags[_] for _ in range(len(extra_flags))),
        out_bin,
    )
    print(src_dir)
    print(cmd)
    proc = subprocess.run(cmd, cwd=src_dir, shell=True, text=True, capture_output=True)
    if proc.returncode != 0:
        raise RuntimeError(
            f"[pygrnwang] Fortran compile failed for {src_dir}\n"
            f"Error:\n{proc.stderr or proc.stdout}"
        )


class CustomBuildPy(_build_py):
    def run(self):
        _build_py.run(self)

        if not shutil.which("gfortran"):
            raise ValueError(
                r"Please install gfortran and ensure "
                r"that command \'gfortran\' can be directly called"
            )
        if not shutil.which("java"):
            raise ValueError(
                r"Please install java and ensure that "
                r"command \'java\' can be directly called"
            )

        exec_dir = os.path.join(project_root, "pygrnwang", "exec")
        os.makedirs(exec_dir, exist_ok=True)

        fortran_src_root = os.path.join(project_root, "fortran_src_codes")
        fortran_subdirs = {
            "edgrn2.0_src": f"edgrn2.{platform_exec}",
            "edcmp2.0_src": f"edcmp2.{platform_exec}",
            "qssp2020_src": f"qssp2020.{platform_exec}",
            "spgrn2020_src": f"spgrn2020.{platform_exec}",
            "qseis06_src": f"qseis06.{platform_exec}",
            "qseis2025_src": f"qseis2025.{platform_exec}",
        }

        for src_folder, bin_name in fortran_subdirs.items():
            fortran_src_dir = os.path.join(fortran_src_root, src_folder)
            output_binary = os.path.join(exec_dir, bin_name)

            extra = []
            env_fflags = os.environ.get("PYGRNWANG_FFLAGS", "")
            if env_fflags:
                extra += env_fflags.split()
            if src_folder == "qseis2025_src":
                extra += ["-ffixed-line-length-none"]

            print(f"[pygrnwang] Compiling {src_folder} -> {output_binary}")
            _compile_dir(fortran_src_dir, output_binary, extra)


setup(cmdclass={"build_py": CustomBuildPy}, distclass=BinaryDistribution)  # type:ignore
